import logging
import sys
from logging import Logger
from logging.handlers import TimedRotatingFileHandler
from os import mkdir, path

from musif.common.constants import (
    LEVEL_CRITICAL,
    LEVEL_DEBUG,
    LEVEL_ERROR,
    LEVEL_INFO,
    LEVEL_WARNING,
)
from musif.common._utils import colorize


class ConsoleFormatter(logging.Formatter):

    format = "%(message)s"

    FORMATS = {
        logging.DEBUG: colorize(format, LEVEL_DEBUG),
        logging.INFO: colorize(format, LEVEL_INFO),
        logging.WARNING: colorize(format, LEVEL_WARNING),
        logging.ERROR: colorize(format, LEVEL_ERROR),
        logging.CRITICAL: colorize(format, LEVEL_CRITICAL),
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class FileFormatter(logging.Formatter):

    FORMAT = "%(asctime)s %(levelname)s %(message)s"

    def format(self, record):
        formatter = logging.Formatter(self.FORMAT)
        return formatter.format(record)


def set_logger_file_handler(
    logger: Logger, log_file_path: str = None, file_log_level: str = None
):
    if log_file_path is None or file_log_level is None:
        return
    log_filename = path.basename(log_file_path)
    log_dir = log_file_path[: -len(log_filename)]
    if not path.exists(log_dir):
        mkdir(log_dir)
    file_handler = TimedRotatingFileHandler(
        log_file_path, when="midnight", backupCount=5
    )
    file_handler.setLevel(file_log_level)
    file_handler.setFormatter(FileFormatter())
    logger.addHandler(file_handler)


def set_logger_console_handler(logger: Logger, console_log_level: str = None):
    if console_log_level is None:
        return
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(console_log_level)
    console_handler.setFormatter(ConsoleFormatter())
    logger.addHandler(console_handler)


def create_logger(
    logger_name: str,
    log_file_path: str = None,
    file_log_level: str = None,
    console_log_level: str = None,
) -> Logger:
    logger = logging.getLogger()
    logger.setLevel(LEVEL_DEBUG)
    logger.propagate = False
    for handler in logger.handlers:
        logger.removeHandler(handler)
    set_logger_file_handler(logger, log_file_path, file_log_level)
    set_logger_console_handler(logger, console_log_level)



from collections import Counter
from itertools import chain
from math import ceil, floor
from typing import List, Tuple, Union

import pandas as pd
import roman
from music21 import pitch, scale
from music21.note import Note
from music21.stream import Measure
from pandas.core.frame import DataFrame

import musif.extract.constants as C
from musif.common.sort import sort_dict
from musif.extract.features.core.handler import DATA_KEY
from musif.extract.features.harmony.utils import get_function_first, get_function_second

accidental_abbreviation = {
    "": "",
    "sharp": "#",
    "flat": "b",
    "double-sharp": "x",
    "double-flat": "bb",
}


def get_keys_functions(list_keys: list, mode: str) -> Tuple[str, str]:
    result_dict = {t: get_function_first(t, mode) for t in set(list_keys)}
    first_function = [result_dict[t] for t in list_keys]
    second_function = [get_function_second(g1) for g1 in first_function]
    return first_function, second_function


def continued_sections(sections: list, mc: dict) -> list:
    extended_sections = []
    repeated_measures = Counter(mc)
    for i, c in enumerate(repeated_measures):
        extended_sections.append([sections[i]] * repeated_measures[c])
    return list(chain(*extended_sections))


def IsAnacrusis(harmonic_analysis):
    return harmonic_analysis.mn.dropna().tolist()[0] == 0


def get_tonality_per_beat(
    harmonic_analysis: DataFrame, tonality: str, start_beat: float, end_beat: float
):
    tonality_map = {}
    for index, degree in enumerate(harmonic_analysis.localkey):
        beat = harmonic_analysis.beats[index]
        tonality_map[beat] = get_localTonalty(tonality, degree.strip())

    _fill_gaps_in_tonality_map(tonality_map, start_beat, end_beat)

    tonality_map = sort_dict(tonality_map, sorted(tonality_map.keys()))
    return tonality_map


def _fill_gaps_in_tonality_map(tonality_map, start_beat, end_beat):
    if 0 not in tonality_map:
        tonality_map[0] = tonality_map[list(tonality_map.keys())[0]]

    for beat in range(floor(start_beat), ceil(end_beat) + 1):
        if beat not in tonality_map:
            if beat-1 in tonality_map:
            # If tonality is not found 
                tonality_map[beat] = tonality_map[beat - 1]
            # In case there is no tonality for first beats, we assume the first one we can find.
            else:
                tonality_map[beat] = tonality_map[list(tonality_map.keys())[0]]


def get_emphasised_scale_degrees_relative(
    notes_list: list, score_data: dict
) -> List[list]:
    harmonic_analysis, tonality = extract_harmony(score_data)
    if harmonic_analysis.size == 0:
        return None

    measures = [
        m for p in score_data[C.DATA_SCORE].parts for m in
        p.getElementsByClass(Measure)
    ]
    min_beat = min(m.offset for m in measures)
    max_beat = max(m.offset + m.quarterLength for m in measures)
    tonality_map = get_tonality_per_beat(
        harmonic_analysis, tonality, min_beat, max_beat
    )
    emph_degrees = get_emphasized_degrees(notes_list, tonality_map, harmonic_analysis)
    return emph_degrees


def get_emphasized_degrees(
    notes_list: List[Note], tonality_map: dict, harmonic_analysis
) -> dict:
    local_tonality = ""
    notes_per_degree_relative = {
        to_full_degree(degree, accidental): 0
        for accidental in ["", "sharp", "flat"]
        for degree in [1, 2, 3, 4, 5, 6, 7]
    }
    for j, note in enumerate(notes_list):
        note = note[0] if note.isChord else note
        if note.measureNumber in list(harmonic_analysis["playthrough"]) and str(note.beat) != 'nan':
            note_offset = round(
                list(harmonic_analysis[harmonic_analysis["playthrough"] == note.measureNumber].beats)[0]
                - 1
                + note.beat
            )
        else:
            note_offset = round(note.offset)

        if note_offset is None:
            note_offset = notes_list[j - 1].offset

        if note_offset in tonality_map:
            local_tonality = tonality_map[round(note_offset)]
        else:
            local_tonality = tonality_map[list(tonality_map.keys())[-1]]

        degree_value = get_note_degree(local_tonality, note.name)

        if str(degree_value) not in notes_per_degree_relative:
            notes_per_degree_relative[str(degree_value)] = 1
        else:
            notes_per_degree_relative[str(degree_value)] += 1

    return notes_per_degree_relative


def get_modulations(lausanne_table: DataFrame, sections, major=True):
    keys = lausanne_table.localkey.dropna().tolist()
    grouping, _ = get_keys_functions(keys, mode="M" if major else "m")
    modulations_sections = {group: [] for group in grouping}
    last_key = ""
    for i, key in enumerate(keys):
        if (key.lower() != "i") and key != last_key:
            section = sections[i]
            last_key = key
            modulation = grouping[i]
            modulations_sections[modulation].append(section)

        if last_key == key and sections[i] != section:
            section = sections[i]
            modulations_sections[modulation].append(section)

    ms = {}
    for m in modulations_sections:
        if len(modulations_sections[m]) != 0:
            ms["Modulations" + str(m)] = len(list(set(modulations_sections[m])))
    return ms


def extract_harmony(score_data):
    harmonic_analysis = score_data.get(C.DATA_MUSESCORE_SCORE, pd.DataFrame())

    tonality = str(score_data[DATA_KEY])

    return harmonic_analysis, tonality


def get_note_degree(key, note):
    if key[0].isupper():
        scl = scale.MajorScale(key.split(" ")[0])
    else:
        scl = scale.MinorScale(key.split(" ")[0])

    degree = scl.getScaleDegreeAndAccidentalFromPitch(pitch.Pitch(note))
    accidental = degree[1].fullName if degree[1] != None else ""

    acc = ""
    if accidental == "sharp":
        acc = "#"
    elif accidental == "flat":
        acc = "b"
    elif accidental == "double-sharp":
        acc = "x"
    elif accidental == "double-flat":
        acc = "bb"

    return acc + str(degree[0])


def get_localTonalty(globalkey, degree):
    accidental = ""
    if "#" in degree:
        accidental = "#"
        degree = degree.replace("#", "")

    elif "b" in degree:
        accidental = "-"
        degree = degree.replace("b", "")

    degree_int = roman.fromRoman(degree.upper())

    if "major" in globalkey:
        pitch_scale = (
            scale.MajorScale(globalkey.split()[0]).pitchFromDegree(degree_int).name
        )
    else:
        pitch_scale = (
            scale.MinorScale(globalkey.split(" ")[0]).pitchFromDegree(degree_int).name
        )

    if ("#" or "b") in [char for char in pitch_scale][
        -1:
    ]:  # checks for exceptions in which we already have an accidental
        modulation = (
            scale.MajorScale(globalkey.split()[0]).pitchFromDegree(degree_int - 1).name
            if "major" in globalkey
            else scale.MinorScale(globalkey.split(" ")[0])
            .pitchFromDegree(degree_int - 1)
            .name
        )

    else:
        modulation = pitch_scale + accidental

    return modulation.upper() if degree.isupper() else modulation.lower()


def to_full_degree(degree: Union[int, str], accidental: str) -> str:
    return f"{accidental_abbreviation[accidental]}{degree}"



"""
This package implements utilities to cache and speed-up the extraction of features.
"""
from .cache import SmartModuleCache, CACHE_FILE_EXTENSION
from .utils import isinstance, iscache, hasattr, store_score_df, FileCacheIntoRAM



from .utils import get_java_path, download_jsymbolic(), JavaNotFoundError

download_jsymbolic()
try:
    get_java_path()
except JavaNotFoundError:
    print("Java not found!")
    print("Value of JAVA_HOME:", os.environ["JAVA_HOME"])
    print("Value of PATH:", os.environ["PATH"])



DYNMEAN = "DynMean"
DYNMEAN_WEIGHTED = "DynMean_weighted"
DYNABRUPTNESS = "DynAbruptness"
DYNGRAD = "DynGrad"

DYNAMIC = "Dynamic"  # all coulf be in a dictionary with this variables?
TEXTEXPRESSION = "TextExpression"
TIMESIGNATURE = "TimeSignature"
REST = "Rest"

DYNAMIC_VALUES = {
    "sfz": 127,
    "fz": 120,
    "ffff": 116,
    "fff": 112,
    "ff": 101,
    "più f": 96,
    "f assai": 94,
    "assaif": 94,
    "f": 88,
    "sempre f": 88,
    "poco f": 80,
    "mf": 75,
    "mp": 62,
    "p": 49,
    "più p": 46,
    "sempre p": 49,
    "p dolce": 49,
    "poco p": 49,
    "p assai": 42,
    "pp": 36,
    "ppp": 24,
    "sfp": 20,
    "pppp": 12,
    "sempre pp": 36,
    "sotto voce assai": 36,
    "dolce": 49,
}

DYNAMIC_LAST_WORD = ["assai", "dolce"]

DYNAMIC_FIRST_WORD = ["più", "sempre", "poco"]



from musif.extract.constants import DATA_PART_ABBREVIATION, DATA_SOUND_ABBREVIATION


def get_part_prefix(part_abbreviation: str) -> str:
    """
    Returns prefix name for a specific part given instrument's abbreviation

    Example
    'vnI' -> 'PartVnI_'

    Parameters
    ----------
    part_abbreviation:  str
        String that represents the abbreviated name of an instrument
    """
    if part_abbreviation is None or len(part_abbreviation) == 0:
        return "Part"
    return f"Part{part_abbreviation[0].upper() + part_abbreviation[1:]}_"


def get_sound_prefix(sound_abbreviation: str) -> str:
    """
    Returns prefix name for a specific part given sound's abbreviation

    Example
    'vnI' -> 'SoundVn_'

    Parameters
    ----------
    part_abbreviation:  str
        String that represents the abbreviated name of a sound
    """
    if sound_abbreviation is None or len(sound_abbreviation) == 0:
        return "Sound"
    return f"Sound{sound_abbreviation[0].upper() + sound_abbreviation[1:]}_"


def get_family_prefix(family_abbreviation: str) -> str:
    """
    Returns prefix name for a specific part given sound's abbreviation

    Example
    'vnI' -> 'SoundVn_'

    Parameters
    ----------
    part_abbreviation:  str
        String that represents the abbreviated name of a sound
    """

    if family_abbreviation is None or len(family_abbreviation) == 0:
        return "Family"
    return f"Family{family_abbreviation[0].upper() + family_abbreviation[1:]}_"


def get_score_prefix() -> str:
    return "Score_"


def get_corpus_prefix() -> str:
    return "Corpus_"


def get_part_feature(part: str, feature: str) -> str:
    """
    It builds the name of a feature with part scope.
    For instance, if the feature is "NumberOfIntervals" and the part is "VnI",
    this class would return: "PartVnI_NumberOfIntervals".
    Args:
        part (str): The part name.
        feature (str): Name of the feature to be prefixed.
    Returns:
        str: The feature properly prefixed for the part passed as argument.
    """

    return get_part_prefix(part) + feature


def get_sound_feature(sound: str, feature: str) -> str:
    """
    It builds the name of a feature with sound scope.

    For instance, if the feature is "NumberOfIntervals" and the sound is "Ob",
    this class would return: "SoundOb_NumberOfIntervals".

    Args:
        sound (str): The sound name.
        feature (str): Name of the feature to be prefixed.

    Returns:
        str: The feature properly prefixed for the sound passed as argument.
    """
    return get_sound_prefix(sound) + feature


def get_family_feature(family: str, feature: str) -> str:
    """
    It builds the name of a feature with family scope.

    For instance, if the feature is "NumberOfIntervals" and the family is "Str",
    this class would return: "FamilyStr_NumberOfIntervals".

    Args:
        family (str): The family name.
        feature (str): Name of the feature to be prefixed.

    Returns:
        str: The feature properly prefixed for the family passed as argument.
    """
    return get_family_prefix(family) + feature


def get_score_feature(feature: str) -> str:
    """
    It builds the name of a feature with Score scope.

    For instance, if the feature is "NumberOfIntervals",
    this class would return: "Score_NumberOfIntervals".

    Args:
        feature (str): Name of the feature to be prefixed.

    Returns:
        str: The feature properly prefixed for score scope.
    """
    return get_score_prefix() + feature



from typing import List, Tuple

from music21.interval import Interval
from music21.note import Note
from music21.scale import MajorScale, MinorScale
from music21.stream.base import Measure, Part, Score, Voice
from music21.text import assembleLyrics
from roman import toRoman

from musif.cache import isinstance


def is_voice(part: Part) -> bool:
    """
    Returns True if the part is a singer part, otherwise returns False

    Parameters
    ----------
    part : Part
      Music21 part to check if it's a singer part
    """

    instrument = part.getInstrument(returnDefault=False)
    if instrument is None or instrument.instrumentSound is None:
        return False
    return "voice" in instrument.instrumentSound


def name_parts(score: Score):
    """
    This function assign a name to each part in the score. If a name is already present,
    we keep it there, otherwise, we create a name of type `missingName#` where # is an
    incremental number,
    """
    i = 0
    for part in score.parts:
        increment = False
        if part.partName is None:
            part.partName = f"NoName{i}"
            increment = True
        if part.partAbbreviation is None:
            part.partAbbreviation = f"NoName{i}"
            increment = True
        if increment:
            i += 1


def split_layers(score: Score, split_keywords: List[str]):
    """
    Function used to split possible layers. Those instruments that include to parts in the same staff
    will be separated in two diferent parts so they can be treated separately.

    Parameters
    ----------
    score : Score
        Music21 score to scan and separate parts in it according to split_keywords list

    split_keywords: List[str]
        List containing key words of instruments susceptible to be splitted. i.e. [oboe, viola]

    """

    final_parts = []
    for part_index, part in enumerate(score.parts):
        instrument = part.getInstrument(returnDefault=False)

        possible_layers = False
        if instrument is not None and instrument.instrumentSound is not None:
            for keyword in split_keywords:
                if keyword in instrument.instrumentSound:
                    possible_layers = True
                    break

        if possible_layers:
            has_voices = False
            for measure in part.elements:
                if isinstance(measure, Measure) and any(
                    isinstance(e, Voice) for e in measure.elements
                ):
                    has_voices = True
                    break

            if (
                has_voices
            ):  # recorrer los compases buscando las voices y separamos en dos parts
                _separate_info_in_two_parts(score, final_parts, part)
            else:
                final_parts.append(part)  # without any change
                score.remove(part)
        else:
            final_parts.append(part)  # without any change
            score.remove(part)

    for part in final_parts:
        try:
            score.insert(0, part)
        except Exception:
            pass


def get_notes_and_measures(
    part: Part,
) -> Tuple[List[Note], List[Note], List[Measure], List[Measure]]:
    """
    Obtains lists of notes, tied notes, measures, measures that containg notes, and notes and rests.
    Information that is useful in the subsequent process of extraction.

    Parameters
    ----------
    part : Part
      Music21 part to extract the info from.

    """

    measures = list(part.getElementsByClass(Measure))
    sounding_measures = [
        idx for idx, measure in enumerate(measures) if len(measure.notes) > 0
    ]
    original_notes = [
        note for measure in measures for note in measure.notes if isinstance(note, Note)
    ]
    notes_and_rests = [n for measure in measures for n in measure.notesAndRests]

    return original_notes, measures, sounding_measures, notes_and_rests


def _separate_info_in_two_parts(score, final_parts, part):
    parts_splitted = part.voicesToParts().elements
    num_measure = 0
    for measure in part.elements:
        # add missing information to both parts (dynamics, text annotations, etc are
        # missing)
        if isinstance(measure, Measure):
            num_measure += 1
            if any(not isinstance(e, Voice) for e in measure.elements):
                not_voices_elements = [
                    e for e in measure.elements if not isinstance(e, Voice)
                ]  # elements such as clefs, dynamics, text annotations...
                for p in parts_splitted:
                    if measure.measureNumber == 0 and isinstance(measure, Measure):
                        # number = measure.measureNumber+1
                        # only add elements if we are in am measure
                        if isinstance(p.elements[num_measure], Measure):
                            p.elements[num_measure].elements += tuple(
                                e
                                for e in not_voices_elements
                                if e not in p.elements[num_measure].elements
                            )
                    if measure.measureNumber > 0:
                        if not isinstance(p.elements[num_measure], Measure):
                            continue
                        p.elements[num_measure].elements += tuple(
                            e
                            for e in not_voices_elements
                            if e not in p.elements[num_measure].elements
                        )
    for num, p in enumerate(parts_splitted, 1):
        p.id = part.id + " " + toRoman(num)  # only I or II
        p.partName = part.partName + " " + toRoman(num)  # only I or II
        # p.elements = p.elements
        final_parts.append(p)

        # p_copy = copy.deepcopy(part)
        # p_copy.id = p_copy.id + " " + toRoman(num)  # only I or II
        # p_copy.partName = p_copy.partName + " " + toRoman(num)  # only I or II
        # p_copy.elements = p.elements
        # final_parts.append(p_copy)
    score.remove(part)


def _get_part_clef(part):
    """
    Extracts the clef in the score by checking the first measure of the part

    Parameters
    ----------
        part : Part
      Music21 part to extract the info from

    """
    for elem in part.elements:
        if isinstance(elem, Measure):
            if hasattr(elem, "clef"):
                return elem.clef.sign + "-" + str(elem.clef.line)
    return ""


def _get_degrees_and_accidentals(key: str, notes: List[Note]) -> List[Tuple[str, str]]:
    if "major" in key.split():
        scl = MajorScale(key.split(" ")[0])
    else:
        scl = MinorScale(key.split(" ")[0])

    degrees = [
        scl.getScaleDegreeAndAccidentalFromPitch(note.pitches[0]) for note in notes
    ]

    return [(degree[0], degree[1].fullName if degree[1] else "") for degree in degrees]


def _get_intervals(notes: List[Note]) -> List[Interval]:
    # TODO: this takes > 6% of the time
    return [
        Interval(notes[i].pitches[0], notes[i + 1].pitches[0])
        for i in range(len(notes) - 1)
    ]


def _contains_text(part: Part) -> bool:
    return assembleLyrics(part)


def _get_lyrics_in_notes(notes: List[Note]) -> List[str]:
    lyrics = []
    for note in notes:
        if note.lyrics is None or len(note.lyrics) == 0:
            continue
        note_lyrics = [
            syllable.text for syllable in note.lyrics if syllable.text is not None
        ]
        lyrics.extend(note_lyrics)
    return lyrics



import itertools
import re
from collections import Counter
from typing import Dict, List

import numpy as np
import pandas as pd
from ms3.expand_dcml import features2type, split_labels
from musif.logs import perr, pwarn
from urllib.request import urlopen
from musif.musicxml.tempo import get_number_of_beats
from musif.logs import perr, pwarn
from .constants import *

REGEX = {}


def get_harmonic_rhythm(ms3_table) -> dict:
    hr = {}
    measures = ms3_table.mn.dropna().tolist()
    measures_compressed = [i for j, i in enumerate(measures) if i != measures[j - 1]]
    numerals = ms3_table.numeral.dropna().tolist()
    number_of_chords = sum(Counter(numerals).values())
    time_signatures = ms3_table.timesig.tolist()
    harmonic_rhythm = (
        number_of_chords / len(measures_compressed)
        if len(measures_compressed) != 0
        else 0.0
    )

    if len(Counter(time_signatures)) == 1:
        harmonic_rhythm_beats = (
            number_of_chords
            / (get_number_of_beats(time_signatures[0]) * len(measures_compressed))
            if len(measures_compressed) != 0
            else 0.0
        )
    else:
        playthrough = ms3_table.playthrough.dropna().tolist()
        periods_ts = []
        time_changes = []

        for t in range(1, len(time_signatures)):
            if time_signatures[t] != time_signatures[t - 1]:

                # what measure in compressed list corresponds to the change in time signature
                time_changes.append(time_signatures[t - 1])
                periods_ts.append(
                    len(measures_compressed[0 : playthrough[t - 1]]) - sum(periods_ts)
                )

        harmonic_rhythm_beats = number_of_chords / sum(
            [
                period * get_number_of_beats(time_changes[j])
                for j, period in enumerate(periods_ts)
            ]
        )

    hr[HARMONIC_RHYTHM] = harmonic_rhythm
    hr[HARMONIC_RHYTHM_BEATS] = harmonic_rhythm_beats

    return hr


def get_measures_per_key(keys_options, measures, keys, mc_onsets, time_signatures):
    key_measures = {p: 0 for p in keys_options}
    last_key = 0
    done = 0
    starting_measure = 0

    new_measures = create_measures_extended(measures)
    numberofmeasures = len(new_measures)

    for i, key in enumerate(keys):
        if key != last_key and i < numberofmeasures:
            n_beats = int(get_number_of_beats(time_signatures[i - 1]))

            if last_key in key_measures:
                num_measures, done = compute_number_of_measures(
                    done,
                    starting_measure,
                    new_measures[i - 1],
                    new_measures[i],
                    mc_onsets[i],
                    n_beats,
                )
                key_measures[last_key] += num_measures

            last_key = key
            starting_measure = new_measures[i] - 1

    # last!
    num_measures, _ = compute_number_of_measures(
        done,
        starting_measure,
        new_measures[numberofmeasures - 1],
        new_measures[numberofmeasures - 1] + 1,
        mc_onsets[numberofmeasures - 1],
        n_beats,
    )

    key_measures[last_key] += num_measures

    try:
        assert not (
            new_measures[0] == 0
            and round(sum(list(key_measures.values()))) != new_measures[i] + 1
        )
    except AssertionError as e:
        perr("There was an error counting the measures!: ", e)
        return {}
    return key_measures


def create_measures_extended(measures):
    new_measures = []
    new_measures.append(measures[0])
    for i in range(1, len(measures)):
        if measures[i] < max(measures[:i]):
            if same_measure(measures, i):
                new_measures.append(new_measures[i - 1])
            else:
                new_measures.append(new_measures[i - 1] + 1)
        else:
            new_measures.append(measures[i])
    return new_measures


def same_measure(measures, i):
    return measures[i] == measures[i - 1]


def compute_number_of_measures(
    done, starting_measure, previous_measure, measure, current_onset, num_beats
):
    starting_measure += done
    if measure == previous_measure:  # We are in the same measure, inside of it

        measures = previous_measure - 1 - starting_measure
        beat = current_onset.numerator
        # habrá que sumarle current_beat / max_beats. Antes convertir current_beat en numérico
        if type(current_onset) == str:
            numbers = current_onset.split(".")
            first = int(numbers[0])
            second = numbers[1].split("/")
            second = int(second[0]) / int(second[1])
            beat = first + second

        return (
            measures + beat / current_onset.denominator,
            beat / current_onset.denominator,
        )
        # before
        # return measures + beat/num_beats, beat/num_beats

    else:
        if (
            measure - previous_measure > 1
        ):  # Change of key happens in a change of measures
            return (
                previous_measure - starting_measure + (measure - 1 - previous_measure),
                0,
            )  ###WTF IS DIS
        else:
            return previous_measure - starting_measure, 0


#########################################################################
# Como sections tiene una indicación por compás, pero a lo largo del script
# trabajamos con la tabla harmonic_analysis, que tiene tantas entradas por
# compás como anotaciones harmónicas, repetimos las secciones según el número


def continued_sections(sections, mc):
    extended_sections = []
    repeated_measures = Counter(mc)
    for i, c in enumerate(repeated_measures):
        extended_sections.append([sections[i]] * repeated_measures[c])

    # Flat list
    return list(itertools.chain(*extended_sections))


################################################################################
# Function to return the harmonic function1 based on the global key mode. Uppercase if
# mode is major, lowercase if minor. 2nd, 4th, adn 6th degrees are considered
# as classes of subdominant function. In major mode, vi is treatred as the relative key (rm);
# in the minor, III = relative major (rj).
# Lowered degrees are indicated with 'b', raised with '#' (bII = Neapolitan key).
# Leading notes are abrreviated as LN.


def get_keys(list_keys, mode):
    result_dict = {t: get_function_first(t, mode) for t in set(list_keys)}
    # result_dict = {t: get_localkey_1(t, mode) for t in set(list_keys)}
    function1 = [result_dict[t] for t in list_keys]
    function2 = [get_function_second(g1) for g1 in function1]
    # function2 = [get_localkey_2(g1) for g1 in function1]
    return function1, function2


###################
# KEYAREAS
###################


def get_keyareas_lists(keys, g1, g2):
    key_areas = []
    key_areas_g1 = []
    key_areas_g2 = []
    last_key = ""
    for i, k in enumerate(keys):
        if k != last_key:
            key_areas.append(k)
            key_areas_g1.append(g1[i])
            key_areas_g2.append(g2[i])
            last_key = k
    return key_areas, key_areas_g1, key_areas_g2


def get_keyareas(lausanne_table, major=True):
    keys = lausanne_table.localkey.dropna().tolist()
    g1, g2 = get_keys(keys, "M" if major else "m")

    key_areas, key_areas_g1, key_areas_g2 = get_keyareas_lists(keys, g1, g2)
    number_blocks_keys = Counter(key_areas)
    # number_blocks_grouping1 = Counter(key_areas_g1)
    # number_blocks_grouping2 = Counter(key_areas_g2)
    measures = lausanne_table.mc.dropna().tolist()
    beats = lausanne_table.mc_onset.dropna().tolist()
    time_signatures = lausanne_table.timesig.tolist()

    key_measures = get_measures_per_key(
        list(set(keys)), measures, keys, beats, time_signatures
    )

    total_measures = float(sum(list(key_measures.values())))
    key_measures_percentage = {
        kc: float(key_measures[kc] / total_measures) for kc in key_measures
    }

    keyGrouping1_measures = get_measures_per_key(
        list(set(g1)), measures, g1, beats, time_signatures
    )
    keyGrouping1_measures = {
        kc: keyGrouping1_measures[kc] / sum(list(keyGrouping1_measures.values()))
        for kc in keyGrouping1_measures
    }
    keyGrouping2_measures = get_measures_per_key(
        list(set(g2)), measures, g2, beats, time_signatures
    )
    keyGrouping2_measures = {
        kc: keyGrouping2_measures[kc] / sum(list(keyGrouping2_measures.values()))
        for kc in keyGrouping2_measures
    }

    total_key_areas = sum(number_blocks_keys.values())
    # total_g1_areas = sum(number_blocks_grouping1.values())
    # total_g2_areas = sum(number_blocks_grouping2.values())

    # keyareas = {'TotalNumberKeyAreas': total_key_areas, 'TotalNumberMeasures': int(total_measures) }
    keyareas = {}
    for key in number_blocks_keys:
        keyareas[KEY_PREFIX + key + KEY_PERCENTAGE] = float(
            key_measures_percentage[key]
        )  # procentaje de compases de cada I, i, etc. en el total
        keyareas[KEY_PREFIX + KEY_MODULATORY + key] = (
            number_blocks_keys[key] / total_key_areas
        )

    return keyareas


def get_function_first(element, mode):
    reference = {
        "T": ["i"],
        "D": ["v", "vii"],
        "SD": ["ii", "iv", "vi"],
        "MED": ["iii"],
    }

    # Special chords
    if any([i for i in ("It", "Ger", "Fr") if i in element]):
        return "D"

    elif element.lower() == "bii":
        return "NAP"

    elif element.lower() in ["#vii", "vii"]:
        return "D"

    if mode == "M":
        if element == "vii":
            return "D"
        elif element == "#vii":
            return "#ln"
        elif element == "bVII":
            return "ST"
        elif element == "bvii":
            return "st"
        elif element == "VII":
            return "LN"

    if mode == "m":
        if element == "#vii":
            return "D"
        elif element == "VII":
            return "ST"
        elif element == "bVII":
            return "bST"
        elif element == "bvii":
            return "bst"
        elif element == "#VII":
            return "LN"
        elif element == "vii":
            return "st"

    element = element.replace("b", "-")  # '-' represents flats
    for key, value in reference.items():
        if element.replace("#", "").replace("-", "").lower() in value:
            output = key.lower() if element.islower() else key
            if "-" in element:
                output = "-" + output
            elif "#" in element:
                output = "#" + output
            return output.replace("-", "b")


def get_function_second(element):
    element = element.replace("b", "-")
    if element.lower() == "#ln":
        return "#ST"
    elif element in ["rm", "rj"]:
        return "rel"
    elif element.upper() in ["ST", "LN"]:
        return "ST"
    else:
        return element.upper().replace("-", "b")


def get_numerals(lausanne_table):
    numerals = lausanne_table.numeral.dropna().tolist()
    keys = lausanne_table.globalkey.dropna().tolist()
    relativeroots = lausanne_table.relativeroot.tolist()

    _, ng2 = get_numerals_lists(
        numerals, relativeroots, keys
    )  # por que se coge solo la funcion segunda?? anyway cojamos los numerals
    numerals_counter = Counter(numerals)

    total_numerals = sum(list(numerals_counter.values()))
    nc = {}
    for n in numerals_counter:
        if str(n) == "":
            raise Exception("Some chords here are not parsed well")
            continue
        nc[NUMERALS_prefix + str(n) + "_Per"] = round(
            (numerals_counter[n] / total_numerals), 3
        )
        nc[NUMERALS_prefix + str(n) + "_Count"] = round((numerals_counter[n]), 3)

    return nc


def get_first_numeral(numeral, relativeroot, local_key):
    if str(relativeroot) != "nan":
        return get_function_first(numeral, "M" if relativeroot.isupper() else "m")
    else:
        return get_function_first(numeral, "M" if local_key.isupper() else "m")


def get_numerals_lists(list_numerals, list_relativeroots, list_local_keys):
    tuples = list(zip(list_numerals, list_relativeroots, list_local_keys))
    result_dict = {t: get_first_numeral(*t) for t in set(tuples)}
    function1 = [result_dict[t] for t in tuples]
    function2 = [get_function_second(g1) for g1 in function1]
    return function1, function2


def get_additions(lausanne_table):
    additions = lausanne_table.changes.tolist()
    total_chords = len(lausanne_table.chord.tolist())
    
    additions_cleaned = []
    for i, a in enumerate(additions):
        if isinstance(a, int):
            additions_cleaned.append(int(a))
        else:
            additions_cleaned.append(str(a))

    additions_counter = Counter(additions_cleaned)
    additions_dict = {
        ADDITIONS_4_6_64: 0,
        ADDITIONS_9: 0,
        OTHERS_NO_AUG: 0,
        OTHERS_AUG: 0,
    }
    for a in additions_counter:
        c = additions_counter[a]
        a = str(a)
        if a == "+9":
            additions_dict[ADDITIONS_9] = c
        elif a in ["4", "6", "64", "4.0", "6.0", "64.0"]:
            additions_dict[ADDITIONS_4_6_64] += c
        elif "+" in a:
            additions_dict[OTHERS_AUG] += c
        elif str(a) == "nan":
            continue
        else:
            additions_dict[OTHERS_NO_AUG] += c

    additions = {}
    for a in additions_dict.keys():
        if additions_dict[a] != 0:
            # additions[ADDITIONS_prefix + str(a)] = additions_dict[a] / sum(
                # list(additions_dict.values())
            additions[ADDITIONS_prefix + str(a)] = additions_dict[a] / total_chords
    return additions


def get_chord_types(lausanne_table):

    chords_forms = make_type_col(
        lausanne_table
    )  # Nan values represent {} notations, not chords

    grouped_forms = get_chord_types_groupings(chords_forms)

    form_counter = Counter(grouped_forms)
    features_chords = {}
    for f in form_counter:
        features_chords[CHORD_TYPES_prefix + str(f)] = form_counter[f] / sum(
            list(form_counter.values())
        )
    return features_chords


def get_chords(harmonic_analysis):

    relativeroots = harmonic_analysis.relativeroot.tolist()
    keys = harmonic_analysis.localkey.dropna().tolist()
    chords = harmonic_analysis.chord.dropna().tolist()
    numerals = harmonic_analysis.numeral.dropna().tolist()
    types = [str(i) for i in harmonic_analysis.chord_type.dropna().tolist()]

    chords_functionalities1, chords_functionalities2 = get_chords_functions(
        chords, relativeroots, keys
    )

    numerals_and_types = [
        str(chord) + str(types[index]).replace("Mm", "").replace("mm", "")
        if types[index] not in ("M", "m")
        else str(chord)
        for index, chord in enumerate(numerals)
    ]

    # chords_order = sort_labels(numerals_and_types, numeral=['I', 'i', 'V', 'v', 'VII', 'vii', 'II', 'ii', 'IV', 'iv','VI','vi','III','iii'], chordtype=['', '7', '+', 'o', '%', 'M', 'm','It'], drop_duplicates=True)

    chords_dict = count_chords(numerals_and_types)  # ,order)

    # Exception for #viio chords
    if "Chord_#viio" in chords_dict:
        chords_dict["Chord_viio"] = (
            chords_dict["Chord_viio"] + chords_dict.pop("Chord_#viio")
            if "Chord_viio" in chords_dict
            else chords_dict.pop("Chord_#viio")
        )

    counter_function_1 = Counter(chords_functionalities1)
    counter_function_2 = Counter(chords_functionalities2)
    chords_group_1 = CountChordsGroup(counter_function_1, "1")
    chords_group_2 = CountChordsGroup(counter_function_2, "2")

    return chords_dict, chords_group_1, chords_group_2


def count_chords(chords: list, order: List[str] = []) -> Dict[str, str]:
    chords_numbers = Counter(chords)
    # chords_numbers=sort_dict(chords_numbers, order)

    total_chords = sum(chords_numbers.values())

    chords_dict = {}
    for degree in chords_numbers:
        chords_dict[CHORD_prefix + degree + "_Per"] = (
            chords_numbers[degree] / total_chords
        )
        chords_dict[CHORD_prefix + degree + "_Count"] = chords_numbers[degree]
    return chords_dict


def CountChordsGroup(counter_function: List[str], number: str) -> Dict[str, str]:
    chords_group = {}
    total_chords_group = sum(Counter(counter_function).values())

    for degree in counter_function:
        chords_group[CHORDS_GROUPING_prefix + number + degree + "_Per"] = (
            counter_function[degree] / total_chords_group
        )
        chords_group[
            CHORDS_GROUPING_prefix + number + degree + "_Count"
        ] = counter_function[degree]

    return chords_group


def parse_chord(chord):
    if "(" in chord:
        chord = chord.split("(")[0]
    if "o" in chord:
        chord = chord.split("o")[0]
    if "+" in chord:
        chord = chord.split("+")[0]
    if "%" in chord:
        chord = chord.split("%")[0]
    if "M" in chord:
        chord = chord.split("M")[0]

    # return chord letter without number
    return re.split("(\d+)", chord)[0]


def get_chord_type(chord_type):
    chord_type = str(chord_type)
    if chord_type == "m":
        return "minor triad"
    elif chord_type == "M":
        return "mayor triad"
    elif chord_type in ["7", "mm7", "Mm7", "MM7", "mM7"]:
        return "7th"
    elif chord_type in ["o", "o7", "%", "%7"]:
        return "dim"
    elif chord_type in ["+", "+M7", "+m7"]:
        return "aug"
    else:
        pwarn(f"Chord type {chord_type} not observed")
        return "other"


def get_chord_types_groupings(chordtype_list):
    return [get_chord_type(chord_type) for chord_type in chordtype_list]


def get_first_chord_local(chord, local_key):
    local_key_mode = "M" if local_key.isupper() else "m"

    if "/" not in chord:
        return get_function_first(parse_chord(chord), local_key_mode)
    else:
        parts = chord.split("/")
        degree = get_function_first(
            parse_chord(parts[0]), "M" if parts[1].isupper() else "m"
        )
        if len(parts) == 2:
            # relative = get_function_first(parts[1], local_key_mode)
            # return '/'.join([degree, relative])
            return degree

        else:
            relative_list = []
            relative_list.append(degree)
            relative_list.append(
                get_function_first(parts[1], "M" if parts[2].isupper() else "m")
            )
            for i in range(2, len(parts)):
                relative_list.append(get_function_first(parts[i], local_key_mode))
            # return '/'.join(relative_list)
            return degree


# Function to return second grouping for any chord in any given local key,
def get_second_grouping_localkey(first_grouping, relativeroot, local_key):
    mode = "M" if local_key else "m"
    if str(relativeroot) != "nan":
        mode = "M" if relativeroot.isupper() else "m"
    parts = first_grouping.split("/")

    degree = get_function_second(parts[0])
    if len(parts) == 2:
        chords = get_function_second(parts[1])  # , mode)
        return degree
        # return '/'.join([degree, chords])

    elif len(parts) == 3:
        # chord_1 = get_degree_2(parts[1], )
        relative_1 = get_function_second(
            parts[1]
        )  # , 'M' if parts[2].isupper() else 'm')
        relative_2 = get_function_second(parts[2])  # , mode)
        return "/".join([degree, relative_1, relative_2])
    return degree


def get_chords_functions(chords, relativeroots, local_keys) -> list:

    chords_localkeys = list(zip(chords, local_keys))
    functionalities_dict = {t: get_first_chord_local(*t) for t in set(chords_localkeys)}

    function_first = [functionalities_dict[t] for t in chords_localkeys]

    # Redefine chords_localkeys to get second chord's functionality
    second_chords_localkeys = list(zip(function_first, relativeroots, local_keys))
    function_second = [
        get_second_grouping_localkey(*first_grouping)
        for first_grouping in second_chords_localkeys
    ]

    return function_first, function_second


def make_type_col(df, num_col="numeral", form_col="form", fig_col="figbass"):
    param_tuples = list(
        df[[num_col, form_col, fig_col]].itertuples(index=False, name=None)
    )
    result_dict = {t: features2type(*t) for t in set(param_tuples)}
    return pd.Series(
        [result_dict[t] for t in param_tuples], index=df.index, name="chordtype"
    ).dropna()


def sort_labels(
    labels, git_branch="master", drop_duplicates=True, verbose=True, **kwargs
):
    """Sort a list of DCML labels following custom criteria.
        Uses: split_labels()
    Parameters
    ----------
    labels : :obj:`collection` or :obj:`pandas.Series`
        The labels you want to sort.
    git_branch : :obj:`str`, optional
        The branch of the DCMLab/standards repo from which you want to use the regEx.
    drop_duplicates : :obj:`bool`, optional
        By default, the function returns an ordered list of unique labels. Set to
        False in order to keep duplicate labels. Note that where the ordered features
        are identical, labels appear in the order of their occurrence.
    verbose : .obj:`bool`, optional
        By default, values that are missing from custom orderings are printed out.
        Pass False to prevent that.
    kwargs : {'values', 'occurrences', 'rvalues', 'roccurrences'}, :obj:`dict`, :obj:`list` or callable
        Pass one argument for every feature that you want to sort in the order
        in which features should be used for sorting. The arguments will be mapped
        on the respective columns which should yield alpha-numeric values to be sorted.
        globalkey
        localkey
        pedal
        chord
        numeral
        form
        figbass
        changes
        relativeroot
        pedalend
        phraseend
        chordtype
    Examples
    --------
    .. highlight:: python
        # Sort numerals by occurrences (descending), the figbass by occurrences (ascending), and
        # the form column by the given order
        sort_labels(labels, numeral='occurrences', figbass='roccurrences', form=['', '+', 'o', '%', 'M'])
        # Sort numerals by custom ordering and each numeral by the (globally) most frequent chord types.
        sort_labels(labels, numeral=['I', 'V', 'IV'], chordtype='occurrences')
        # Sort relativeroots alphabetically and the numerals by a custom ordering which
        # is equivalent to ['V', 'vii', '#vii']
        sort_labels(labels, relativeroot='rvalues', numeral={'vii': 5, 'V': 0,  '#vii': 10})
        # Sort chord types by occurrences starting with the least frequent and sort their inversions
        # following the given custom order
        sort_labels(labels, chordtype='roccurrences', figbass=['2', '43', '65', '7'])
    """
    if len(kwargs) == 0:
        raise ValueError("Pass at least one keyword argument for sorting...")
    if not isinstance(labels, pd.Series):
        if isinstance(labels, pd.DataFrame):
            raise TypeError("Pass only one column of your DataFrame.")
        labels = pd.Series(labels)
    if drop_duplicates:
        labels = labels.drop_duplicates()
    features = split_labels(labels, git_branch)

    def make_keys(col, order):
        def make_order_dict(it):
            missing = [v for v in col.unique() if v not in it]
            if len(missing) > 0 and verbose:
                print(
                    f"The following values were missing in the custom ordering for column {col.name}:\n{missing}"
                )
            return {v: i for i, v in enumerate(list(it) + missing)}

        if order in ["values", "rvalues"]:
            keys = sorted(set(col)) if order == "values" else reversed(sorted(set(col)))
            order_dict = make_order_dict(keys)
        elif order in ["occurrences", "roccurrences"]:
            keys = (
                col.value_counts(dropna=False).index
                if order == "occurrences"
                else col.value_counts(dropna=False, ascending=True).index
            )
            order_dict = make_order_dict(keys)
        elif order.__class__ is not dict:
            try:
                order_dict = make_order_dict(order)
            except:
                # order is expected to be a callable:
                return np.vectorize(order)(col)
        else:
            order_dict = order

        return np.vectorize(order_dict.get)(col)

    if "chordtype" in kwargs:
        features["chordtype"] = make_type_col(features)
    key_cols = {
        col: make_keys(features[col], order)
        for col, order in kwargs.items()
        if col in features.columns
    }
    df = pd.DataFrame(key_cols, index=features.index)
    ordered_ix = df.sort_values(by=df.columns.to_list()).index
    return labels.loc[ordered_ix]


def split_labels(labels, git_branch="master", dropna=True):
    """Split DCML harmony labels into their respective features using the regEx
        from the indicated branch of the DCMLab/standards repository.
    Parameters
    ----------
    labels : :obj:`pandas.Series`
        Harmony labels to be split.
    git_branch : :obj:`str`, optional
        The branch of the DCMLab/standards repo from which you want to use the regEx.
    dropna : :obj:`bool`, optional
        Drop rows where the regEx didn't match.
    """
    global REGEX
    if git_branch not in REGEX:
        url = f"https://raw.githubusercontent.com/DCMLab/standards/{git_branch}/harmony.py"
        glo, loc = {}, {}
        exec(urlopen(url).read(), glo, loc)
        REGEX[git_branch] = re.compile(loc["regex"], re.VERBOSE)
    regex = REGEX[git_branch]
    cols = [
        "globalkey",
        "localkey",
        "pedal",
        "chord",
        "numeral",
        "form",
        "figbass",
        "changes",
        "relativeroot",
        "pedalend",
        "phraseend",
    ]
    res = labels.str.extract(regex, expand=True)[cols]
    if dropna:
        return res.dropna(how="all").fillna("")
    return res.fillna("")



import itertools
from typing import Union

import ms3
import music21 as m21
import pandas as pd
from music21.stream.base import Measure
from pandas import DataFrame

import musif.extract.constants as C
from musif.cache import isinstance
from musif.extract.constants import PLAYTHROUGH
from musif.logs import pwarn
from musif.musicxml.tempo import get_number_of_beats

file_names = []
repeat_bracket = False


def process_musescore_file(file_path: str, expand_repeats: bool = False) -> DataFrame:
    """
    Given a mscx file name, parses the file using ms3 library and returns a dataframe containing all harmonic information.
    Adds Playthrough column that contains number of every measure in the cronological order
    Parameters
    ----------
    file_path: str
        Path to mscx file
    expand_repeats: bool
        Directory path to musescore file
    Returns
    -------
    harmonic_analysis: str
        Dataframe containing harmonic information
    """

    msc3_score = ms3.score.Score(file_path, logger_cfg={"level": "ERROR"})
    harmonic_analysis = msc3_score.mscx.expanded()
    harmonic_analysis.reset_index(inplace=True)
    if expand_repeats:
        harmonic_analysis = msc3_score.mscx.expanded(unfold=True)
        harmonic_analysis.reset_index(inplace=True)
        # unfolded_mc=msc3_score.mscx.measures().set_index("mc").next
        # mn = next2sequence(unfolded_mc)
        # mn = ms3.utils.next2sequence(unfolded_mc)
        # mn = pd.Series(mn, name="mc_playthrough")
        # harmonic_analysis = ms3.utils.unfold_repeats(harmonic_analysis, mn)
        harmonic_analysis.rename(columns={"mc_playthrough": PLAYTHROUGH}, inplace=True)
    else:
        harmonic_analysis = msc3_score.mscx.expanded()
        harmonic_analysis.reset_index(inplace=True)
        if harmonic_analysis.mn[0] == 0:
            harmonic_analysis[PLAYTHROUGH] = harmonic_analysis["mc"]
        else:
            harmonic_analysis[PLAYTHROUGH] = harmonic_analysis["mn"]
    _include_beats_column(harmonic_analysis)
    return harmonic_analysis


def expand_score_repetitions(score, repeat_elements: list):
    """
    Given a music21 Score object and a list containing repetition elements, expands the score object and
    places all measures in their correspondent cronological order
    Parameters
    ----------
    score: music21 Score
        Score object parsed by music21
    expand_repeats: list
        List containing all repetition elements
    Returns
    -------
    final_score: music21 Score
        Score object with expanded repetitions
    """

    score = _expand_repeat_bars(score)
    final_score = m21.stream.Score()
    final_score.metadata = score.metadata

    if len(repeat_elements) > 0:
        for part in score.parts:
            p = _expand_part(part, repeat_elements)

            final_score.insert(0, p)
    else:
        final_score = score
    return final_score


def _measure_ranges(
    instrument_measures: int,
    init: int,
    end: int,
    iteration: int = None,
    offset: int = None,
    twocompasses_flag: bool = False,
    remove_repetition_marks_flag: bool = False,
) -> list:
    measures = []
    o = offset
    last_offset = (
        0.0 if int(init) - 6 < 0 else instrument_measures[int(init) - 6].offset
    )

    init_index, end_index = _find_init_and_end_indexes(instrument_measures, init, end)
    for i in range(init_index, end_index + 1):
        if (
            not i < 0
            and i < len(instrument_measures)
            and instrument_measures[i].measureNumber >= int(init)
            and instrument_measures[i].measureNumber <= int(end)
        ):
            if not twocompasses_flag:
                compass = instrument_measures[i].quarterLength
                m = m21.stream.Measure(number=instrument_measures[i].measureNumber)
                if remove_repetition_marks_flag:
                    m.elements = [
                        e
                        for e in instrument_measures[i].elements
                        if not isinstance(e, m21.repeat.RepeatMark)
                    ]
                else:
                    m.elements = instrument_measures[i].elements
                m.quarterLength = compass
                if offset is None:
                    m.offset = last_offset
                else:
                    m.offset = o
                    o += compass

                measures.append(m)
                if (
                    instrument_measures[i].measureNumber != 0.0
                    and instrument_measures[i].offset != 0
                ):
                    last_offset = instrument_measures[i].offset + compass
            twocompasses_flag = False

    if iteration == 2:
        last_measure = instrument_measures[i + 1]
        last_measure.offset = measures[-1].offset
        measures = measures[:-1] + [last_measure]
    return measures


def _find_init_and_end_indexes(
    instrument_measures: list, init: int, end: int
) -> Union[int, int]:
    init_index = instrument_measures.index(
        [m for m in instrument_measures if m.measureNumber == init][0]
    )
    end_compass = [m for m in instrument_measures if m.measureNumber == end]
    end_index = (
        instrument_measures.index(end_compass[0])
        if len(end_compass) > 0
        else len(instrument_measures) - 1
    )
    return init_index, end_index


def _get_instrument_elements(part):
    measures = []
    for elem in part:
        if isinstance(elem, m21.stream.Measure):
            measures.append(elem)
            # Change the note offsets to avoid problems due to the slurs
            last_offset = 0
            last_duration = 0
            for note in elem:
                if (
                    isinstance(note, m21.note.Note)
                    or isinstance(note, m21.note.Rest)
                    or isinstance(note, m21.chord.Chord)
                ):
                    note.offset = last_offset + last_duration
                    last_offset = note.offset
                    last_duration = note.duration.quarterLength
    return measures


def get_repetition_elements(score, v=True):
    global repeat_bracket
    repeat_bracket = False
    repeat_elements = set()

    for instruments in score.parts:
        instr_repeat_elements = []
        for elem in instruments.elements:
            if isinstance(elem, m21.stream.Measure):
                for e in elem:
                    if isinstance(e, m21.repeat.RepeatMark) and not isinstance(
                        e, m21.bar.Repeat
                    ):
                        measure = e.measureNumber
                        if elem.numberSuffix in ["X1", "X2"]:  # Exception
                            measure += 1
                        instr_repeat_elements.append((measure, e.name))
            elif isinstance(elem, m21.spanner.RepeatBracket):
                repeat_bracket = True
                string_e = str(elem)
                index = string_e.find("music21.stream.Measure")
                string_e = string_e[index:].replace("music21.stream.Measure ", "")
                measure = (
                    string_e.split(" ")[0].strip().replace("X1", "").replace("X2", "")
                )
                instr_repeat_elements.append(
                    (int(measure), "repeat bracket" + elem.number)
                )
        repeat_elements.update(instr_repeat_elements)

    repeat_elements = sorted(list(repeat_elements), key=lambda tup: tup[0])
    if v:
        print("The repeat elements found in this score are: " + str(repeat_elements))
    return repeat_elements


def _include_beats_column(harmonic_analysis: DataFrame) -> None:
    harmonic_analysis["beats"] = 0
    for index, measure in enumerate(harmonic_analysis[PLAYTHROUGH].values):
        if measure <= 1:
            beat = int(
                measure
                + float(harmonic_analysis.mc_onset[index])
                * get_number_of_beats(harmonic_analysis.timesig[index])
            )
        else:
            time_sig = get_number_of_beats(harmonic_analysis.timesig[index - 1])
            beat = int(
                (measure - 1) * time_sig
                + 1
                + harmonic_analysis.mc_onset[index]
                * get_number_of_beats(harmonic_analysis.timesig[index])
            )

        harmonic_analysis.at[index, "beats"] = beat


def _get_beat_position(
    beats_timesignature: int, number_of_beats: int, pos: int
) -> float:
    if number_of_beats == beats_timesignature:
        return pos
    else:
        return (pos / beats_timesignature) + 1


def _expand_repeat_bars(score):
    final_score = m21.stream.Score()
    final_score.metadata = score.metadata
    exist_repetition_bars = False

    # find repeat bars and expand
    for instr in score.parts:
        part_measures = _get_instrument_elements(
            instr.elements
        )  # returns the measures with repetitions
        last_measure = part_measures[-1].measureNumber
        part_measures_expanded = []
        startsin0 = part_measures[0].measureNumber == 0  # Everything should be -1
        repetition_bars = []

        # Find all repetitions in that instrument
        for elem in instr.elements:
            if isinstance(elem, m21.stream.Measure):
                exist_repetition_bars = _examine_measure(repetition_bars, elem)
            elif isinstance(elem, m21.spanner.RepeatBracket):
                _examine_repeat_bracket(instr, repetition_bars, elem)
        repetition_bars = sorted(list(repetition_bars), key=lambda tup: tup[0])

        start = 0 if startsin0 else 1
        _append_repetitions(
            final_score,
            exist_repetition_bars,
            instr,
            part_measures,
            last_measure,
            part_measures_expanded,
            repetition_bars,
            start,
        )

    return final_score if exist_repetition_bars else score


def _append_repetitions(
    final_score,
    exist_repetition_bars,
    instr,
    part_measures,
    last_measure,
    part_measures_expanded,
    repetition_bars,
    start,
):
    if exist_repetition_bars:
        p = m21.stream.Part()
        p.id = instr.id
        p.partName = instr.partName
        for repetition_bar in repetition_bars:
            measure = _measure_ranges(
                part_measures, repetition_bar[0], repetition_bar[0]
            )[0].quarterLength
            if repetition_bar[1] == "start":
                start = _add_start(
                    part_measures,
                    part_measures_expanded,
                    start,
                    repetition_bar,
                    measure,
                )
            elif repetition_bar[1] == "end":
                start = _add_end(
                    part_measures,
                    part_measures_expanded,
                    repetition_bars,
                    start,
                    repetition_bar,
                    measure,
                )
        if start < last_measure:
            measure = _measure_ranges(part_measures, start, start + 1)[0].quarterLength
            offset = part_measures_expanded[-1][-1].offset
            part_measures_expanded.append(
                _measure_ranges(
                    part_measures, start, last_measure + 1, offset=offset + measure
                )
            )

        p.elements = list(itertools.chain(*tuple(part_measures_expanded)))
        final_score.insert(0, p)


def _add_end(
    part_measures,
    part_measures_expanded,
    repetition_bars,
    start,
    repetition_bar,
    measure,
):
    if len(part_measures_expanded) > 0:
        offset = part_measures_expanded[-1][-1].offset
    else:
        offset = 0
    casilla_1 = (
        True
        if any(re[1] == "1" and re[0] <= repetition_bar[0] for re in repetition_bars)
        else False
    )
    casilla_2 = None
    if casilla_1:
        casilla_2 = [
            re for re in repetition_bars if re[1] == "2" and re[0] > repetition_bar[0]
        ]
        casilla_2 = None if len(casilla_2) == 0 else casilla_2[0]
    part_measures_expanded.append(
        _measure_ranges(
            part_measures,
            init=start,
            end=repetition_bar[0],
            offset=offset + measure,
            remove_repetition_marks_flag=True,
        )
    )  # This should erase the repetition marks
    if casilla_2 != None:
        part_measures_expanded.append(
            _measure_ranges(
                part_measures,
                start,
                casilla_2[0],
                iteration=2,
                offset=part_measures_expanded[-1][-1].offset + measure,
            )
        )
        start = casilla_2[0] + 1
    else:
        part_measures_expanded.append(
            _measure_ranges(
                part_measures,
                init=start,
                end=repetition_bar[0],
                offset=part_measures_expanded[-1][-1].offset + measure,
            )
        )
        start = repetition_bar[0] + 1
    return start


def _add_start(part_measures, part_measures_expanded, start, repetition_bar, measure):
    if len(part_measures_expanded) > 0:
        offset = part_measures_expanded[-1][-1].offset
    else:
        offset = 0
    start_measures = _measure_ranges(
        part_measures, start, repetition_bar[0] - 1, offset=offset + measure
    )
    if len(start_measures) > 0:
        part_measures_expanded.append(start_measures)
    start = repetition_bar[0]
    return start


def _examine_repeat_bracket(instr, repetition_bars, elem):
    string_e = str(elem)
    index = string_e.find("music21.stream.Measure")
    measure = string_e[index:].replace("music21.stream.Measure", "")[1:3].strip()
    repetition_bars.append((int(measure), elem.number))
    index = instr.elements.index(elem)
    elem.elements = instr.elements[:index] + instr.elements[index + 1 :]


def _examine_measure(repetition_bars, elem):
    for e in elem:
        if isinstance(e, m21.bar.Repeat):
            exist_repetition_bars = True
            if e.direction == "start":
                repetition_bars.append((e.measureNumber, "start"))
            elif e.direction == "end":
                repetition_bars.append((e.measureNumber, "end"))
            index = elem.elements.index(e)
            elem.elements = elem.elements[:index] + elem.elements[index + 1 :]
    return exist_repetition_bars


def _get_measures_list(part_measures: list, repeat_elements: list):
    measures_list = []
    startsin0 = part_measures[0].measureNumber == 0  # Everything should be -1

    there_is_fine = False
    there_is_segno = False
    # 1. find the fine and segno
    if any([r[1] == "fine" for r in repeat_elements]):
        f = [x[0] for x in repeat_elements if x[1] == "fine"][0]
        there_is_fine = True
    if any([r[1] == "segno" for r in repeat_elements]):
        s = [x[0] for x in repeat_elements if x[1] == "segno"][0]
        there_is_segno = True

    # Having all the repetition elements, get the measures
    if there_is_segno:
        before_segno = _measure_ranges(part_measures, 1 if not startsin0 else 0, s - 1)
        measures_list.append(
            before_segno
        )  # S -1 OR S-> when segno in compass 1, s, else s-1?
        dc_time_signature = [
            y
            for x in before_segno
            for y in x.elements
            if isinstance(y, m21.meter.TimeSignature)
        ]
    elif there_is_fine:
        measures_list.append(
            _measure_ranges(
                part_measures,
                1 if not startsin0 else 0,
                f - 1,
                iteration=1 if repeat_bracket else None,
            )
        )
    else:
        measures_list.append(
            _measure_ranges(
                part_measures,
                1 if not startsin0 else 0,
                len(part_measures),
                iteration=1 if repeat_bracket else None,
            )
        )

    for repeat in repeat_elements:
        repeat_measure = _measure_ranges(part_measures, repeat[0], repeat[0])
        compass = repeat_measure[0].quarterLength

        if repeat[1] == "segno":
            offset = measures_list[-1][-1].offset
            segno_part = _measure_ranges(
                part_measures,
                s,
                f - 1 if there_is_fine else len(part_measures),
                iteration=1 if repeat_bracket else None,
                offset=offset + compass,
                remove_repetition_marks_flag=True,
            )
            measures_list.append(segno_part)  # Segno to Fine
        elif repeat[1] == "fine":
            twoCompasses = False
            """if len(repeat_measure) > 0:
                twoCompasses = True"""
            offset = measures_list[-1][-1].offset
            fine_part = _measure_ranges(
                part_measures,
                f,
                len(part_measures),
                offset=offset + compass,
                twocompasses_flag=twoCompasses,
                remove_repetition_marks_flag=True,
            )
            measures_list.append(fine_part)  # Fine to end
        elif repeat[1] == "al segno" or repeat[1] == "dal segno":
            offset = measures_list[-1][-1].offset
            # segnos' compass time signature
            segno_time_measure = [
                x
                for x in segno_part[0].elements
                if isinstance(x, m21.meter.TimeSignature)
            ]
            segno_time_measure = (
                segno_time_measure
                if len(segno_time_measure) != 0
                else dc_time_signature[-1]
            )
            alsegno_list = _measure_ranges(
                part_measures,
                s,
                f - 1 if there_is_fine else len(part_measures),
                iteration=2 if repeat_bracket else None,
                offset=offset + compass,
                remove_repetition_marks_flag=True,
            )
            if not any(
                isinstance(x, m21.meter.TimeSignature) for x in alsegno_list[0].elements
            ):
                # we reset the time signature that was on the dacapo
                alsegno_list[0].elements = tuple(
                    [segno_time_measure] + list(alsegno_list[0].elements)
                )

            measures_list.append(alsegno_list)  # Segno to fine
        elif repeat[1] == "da capo":
            offset = measures_list[-1][-1].offset
            if startsin0 and there_is_fine and not repeat_bracket:
                f += 1
            dacapo_list = _measure_ranges(
                part_measures,
                0 if startsin0 else 1,
                f - 1 if there_is_fine else len(part_measures),
                iteration=2 if repeat_bracket else None,
                offset=offset + compass,
                remove_repetition_marks_flag=True,
            )
            measures_list.append(dacapo_list)

    return measures_list


def _expand_part(part, repeat_elements: list):
    part_measures = _get_instrument_elements(
        part.elements
    )  # returns the measures with repetitions
    p = m21.stream.Part()
    p.id = part.id
    p.partName = part.partName

    part_measures_expanded = _get_measures_list(
        part_measures, repeat_elements
    )  # returns the measures expanded
    part_measures_expanded = list(itertools.chain(*part_measures_expanded))
    # Assign a new continuous measure number to every measure
    measure_number = 0 if part_measures_expanded[0].measureNumber == 0 else 1
    for i, e in enumerate(part_measures_expanded):
        m = m21.stream.Measure(number=measure_number)
        m.elements = e.elements
        m.offset = e.offset
        m.quarterLength = e.quarterLength
        part_measures_expanded[i] = m
        measure_number += 1
    p.elements = part_measures_expanded
    return p


def _calculate_total_number_of_beats(time_signatures: list) -> int:
    """
    Given a list of time signatures, sums the beats of each time signature
    """
    # periods = _get_timesignature_periods(time_signatures)
    return sum(get_number_of_beats(ts) for ts in time_signatures)


def cast_mixed_dtypes(col):
    if "mixed" in pd.api.types.infer_dtype(col):
        notna = col.notna()
        newtype = col[notna].map(type).mode()[0]
        if issubclass(newtype, float):
            #  convert fractions like '1/3' to float
            col[notna] = col[notna].apply(pd.eval)
            col = col.convert_dtypes()
        elif issubclass(newtype, int):
            # convert to string
            col = col.astype("string")
    return col


def extract_global_time_signature(score_data):
    """
    Extracts a global time signature for the score for cases where is not possibel to get measure-by-measure TS
    """
    global_ts = (
        score_data[C.DATA_FILTERED_PARTS][0]
        .getElementsByClass(Measure)[0]
        .timeSignature
    )
    if global_ts is None:
        pwarn("Couldn't extract Global Time Sigature!")
        global_ts = "NA"
    score_data[C.GLOBAL_TIME_SIGNATURE] = global_ts
    return global_ts



from pathlib import PurePath

from musif.common._logs import create_logger
from musif.common._utils import read_object_from_yaml_file
from musif.extract.constants import REQUIRE_MSCORE

# TODO: add documentation for these variables
LOGGER_NAME = "musif"
LOG = "log"
LOG_FILE_PATH = "log_file"
FILE_LOG_LEVEL = "file_log_level"
CONSOLE_LOG_LEVEL = "console_log_level"
XML_DIR = "xml_dir"
MUSESCORE_DIR = "musescore_dir"
CACHE_DIR = "cache_dir"
IGNORE_ERRORS = "ignore_errors"
PARALLEL = "parallel"
FEATURES = "features"
BASIC_MODULES = "basic_modules"
BASIC_MODULES_ADDRESSES = "basic_modules_addresses"
FEATURE_MODULES_ADDRESSES = "feature_modules_addresses"
PARTS_FILTER = "parts_filter"
EXPAND_REPEATS = "expand_repeats"
WINDOW_SIZE = "window_size"
OVERLAP = "overlap"
PRECACHE_HOOKS = "precache_hooks"
REMOVE_UNPITCHED_OBJECTS = "remove_unpitched_objects"
MSCORE_EXEC = "mscore_exec"
JSYMBOLIC_CONFIG_FILE = "jsymbolic_config_file"
JSYMBOLIC_MAX_RAM = "jsymbolic_max_ram"
# Didone specific?
SPLIT_KEYWORDS = "split_keywords"

DELETE_FILES = "delete_failed_files"
DELETE_HARMONY = "delete_files_without_harmony"
UNBUNDLE_INSTRUMENTATION = "separate_intrumentation_column"
INSTRUMENTS_TO_KEEP = "instruments_to_keep"
INSTRUMENTS_TO_DELETE = "instruments_to_delete"
ENDSWITH = "columns_endswith"
STARTSWITH = "columns_startswith"
CONTAIN = "columns_contain"
MATCH = "columns_match"
REPLACE_NANS = "replace_nans"
DFS_DIR = "dfs_dir"
GROUPED = "grouped_analysis"
MERGE_VOICES = "merge_voices"
MAX_NAN_COLUMNS = "max_nan_columns"
MAX_NAN_ROWS = "max_nan_rows"

_CONFIG_LOG_FALLBACK = {
    LOG_FILE_PATH: "./musif.log",
    FILE_LOG_LEVEL: "DEBUG",
    CONSOLE_LOG_LEVEL: "ERROR",
}

_CONFIG_FALLBACK = {
    XML_DIR: None,
    MUSESCORE_DIR: None,
    CACHE_DIR: None,
    PARALLEL: 1,
    PRECACHE_HOOKS: [],
    BASIC_MODULES: [],
    IGNORE_ERRORS: False,
    BASIC_MODULES_ADDRESSES: ["musif.extract.basic_modules"],
    FEATURE_MODULES_ADDRESSES: ["musif.extract.features"],
    FEATURES: ["core"],
    SPLIT_KEYWORDS: [],
    PARTS_FILTER: [],
    EXPAND_REPEATS: False,
    WINDOW_SIZE: None,
    OVERLAP: 2,
    MSCORE_EXEC: None,
    DFS_DIR: None,
    REMOVE_UNPITCHED_OBJECTS: True,
    JSYMBOLIC_MAX_RAM: "4g",
    JSYMBOLIC_CONFIG_FILE: None
}

_CONFIG_POST_FALLBACK = {
    DELETE_FILES: False,
    GROUPED: False,
    DELETE_FILES: False,
    DELETE_HARMONY: False,
    UNBUNDLE_INSTRUMENTATION: False,
    MERGE_VOICES: True,
    INSTRUMENTS_TO_KEEP: [],
    INSTRUMENTS_TO_DELETE: [],
    ENDSWITH: [],
    STARTSWITH: [],
    CONTAIN: [],
    REPLACE_NANS: [],
    MATCH: [],
    MAX_NAN_COLUMNS: None,
    MAX_NAN_ROWS: None
}


class GenericConfiguration:
    """
    Generic class for configuration objects.

    When subclassing, you should override the `_get_fallback` method.
    """

    def _get_fallback(self):
        """
        Returns a dictionary containing the default values for this configuration
        object. Such values will be always available, but can be overriden by the
        user.
        """
        return {}

    def __init__(self, arg, **kwargs):
        """
        It set up the `Configuration` object in this way:

            #. it sets the default values returned by `self._get_fallback()` as filed of
               the object (i.e. in the `__dict__` dictionary)
            #. it load the configuration file `arg` provided by the user
            #. it overrides the configuration file with the keyword arguments
            #. it sets the values obtained, overriding the one already available, in
               `__dict_`, making them avilable as usual fields
        """
        # set default values
        for fallback in [_CONFIG_LOG_FALLBACK, self._get_fallback()]:
            for k, v in fallback.items():
                self.__dict__[k] = v

        # load configuration file
        config_data = {}
        if arg is not None:
            if isinstance(arg, str) or isinstance(arg, PurePath):
                config_data = read_object_from_yaml_file(arg)
            elif isinstance(arg, dict):
                config_data = arg
            elif isinstance(arg, GenericConfiguration):
                config_data = arg.to_dict()
            else:
                raise TypeError(
                    f"The argument type is {type(arg)}, and it was expected a dictionary, a Configuration or a string object"
                )

        # override values with kwargs
        config_data.update(kwargs)
        log_config = config_data.get(LOG, {})
        for config in [config_data, log_config]:
            for k, v in config.items():
                if k == LOG:
                    continue
                self.__dict__[k] = v

        create_logger(
            LOGGER_NAME, self.log_file, self.file_log_level, self.console_log_level
        )

    def to_dict(self) -> dict:
        """
        Returns a dictionary having as keys the public fields of this object.
        """
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}


class ExtractConfiguration(GenericConfiguration):
    """
    Configuration object used by :class: `musif.extract.extract.FeatureExtractor`

    It additionaly sets the following properties:

    ..  code-block:: python

        from musif.musicxml import constants as musicxml_c
        self.scoring_family_order = musicxml_c.SCORING_FAMILY_ORDER
        self.scoring_order = musicxml_c.SCORING_ORDER
        self.sound_to_family = musicxml_c.SOUND_TO_FAMILY
        self.family_to_abbreviation = musicxml_c.FAMILY_TO_ABBREVIATION
        self.sound_to_abbreviation = musicxml_c.SOUND_TO_ABBREVIATION

    The above settings can be overriden by the user both by changing
    the variables in `musicxml.constants` and by adding them to the configuration.
    """

    def __init__(self, *args, **kwargs):
        from musif.musicxml import constants as musicxml_c
        # ^--- here to avoid circular imports
        self.scoring_family_order = musicxml_c.SCORING_FAMILY_ORDER
        self.scoring_order = musicxml_c.SCORING_ORDER
        self.sound_to_family = musicxml_c.SOUND_TO_FAMILY
        self.family_to_abbreviation = musicxml_c.FAMILY_TO_ABBREVIATION
        self.sound_to_abbreviation = musicxml_c.SOUND_TO_ABBREVIATION
        super().__init__(*args, **kwargs)

    def _get_fallback(self):
        return _CONFIG_FALLBACK

    def is_requested_musescore_file(self) -> bool:
        """
        Returns `True` if any of the requested features needs musescore files for the
        harmonic annotations.
        Returns `False` otherwise.
        """

        for feature in REQUIRE_MSCORE:
            if self.is_requested_feature_category(feature):
                return True
        return False

    def is_requested_feature_category(self, feature: str) -> bool:
        """
        Returns `True` if `feature` is among the requeste features, according to the
        configuration.
        Returns `False` otherwise.
        """

        if self.features is None:
            return True
        return feature in self.features


class PostProcessConfiguration(GenericConfiguration):
    """
    Configuration object used by :class: `musif.process.DataProcessor`
    """

    def _get_fallback(self):
        return _CONFIG_POST_FALLBACK



from copy import deepcopy
from statistics import mean
from typing import List
from xml.dom.minidom import Element

from musif.config import ExtractConfiguration
from musif.extract.features.prefix import get_part_feature, get_score_feature
from musif.extract.utils import _get_beat_position
from musif.logs import lwarn, pwarn
from musif.musicxml.tempo import get_number_of_beats

from ...constants import DATA_PART_ABBREVIATION
from .constants import *


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    dynamics = []
    beats_section = 0
    dyn_mean_weighted = 0
    total_beats = 0
    total_sounding_beats = 0
    dyn_grad = 0
    last_dyn = 0
    number_of_beats = 1
    name = ""
    old_beat = 0
    dyn = False
    first_silence = False
    beats_timesignature = get_number_of_beats("4/4")

    for measure in part_data["measures"]:
        measure_elements = list(measure.elements)
        for element in measure_elements:
            if (
                element.classes[0] == DYNAMIC
                and not element.value == "sf"
                or (
                    element.classes[0] == TEXTEXPRESSION
                    and element.content == ("sotto voce assai" and "dolce")
                )
            ):
                if element.classes[0] == TEXTEXPRESSION:
                    name += element.content
                else:
                    if element.value in DYNAMIC_FIRST_WORD:
                        name = element.value + name
                    elif name.strip() in DYNAMIC_LAST_WORD:
                        name = element.value + name
                    else:
                        name += element.value
                dyn = True
            elif element.classes[0] == TEXTEXPRESSION:
                if element.content in DYNAMIC_LAST_WORD:
                    name += " " + element.content
                elif element.content in DYNAMIC_FIRST_WORD:
                    name = element.content + " "
            elif element.classes[0] == TIMESIGNATURE:
                beats_timesignature = element.beatCount
                number_of_beats = get_number_of_beats(element.ratioString)
            elif element.classes[0] == REST and not first_silence:
                if element.duration.quarterLength >= number_of_beats:
                    first_silence = True
                    new_dyn = 0
                    dynamics.append(new_dyn)
                    position = _get_beat_position(
                        beats_timesignature, number_of_beats, element.beat
                    )
                    old_beat = position - _get_beat_position(
                        beats_timesignature, number_of_beats, 1
                    )
                    dyn_mean_weighted += (beats_section + old_beat) * last_dyn
                    beats_section, dyn_grad, last_dyn = calculate_gradient(
                        beats_section, dyn_grad, last_dyn, old_beat, new_dyn
                    )
                    name = ""
            if dyn:
                if name in ["fp", "pf"]:
                    new_dyn = get_dynamic_numeric(name[0])
                    if new_dyn != last_dyn:
                        (
                            beats_section,
                            dyn_grad,
                            last_dyn,
                            first_silence,
                            dyn_mean_weighted,
                        ) = calculate_dynamics(
                            dynamics,
                            beats_section,
                            dyn_mean_weighted,
                            dyn_grad,
                            last_dyn,
                            number_of_beats,
                            element,
                            beats_timesignature,
                            new_dyn,
                        )
                    name = name[1]
                if name == "other-dynamics":
                    file_name = score_data["file"]
                    lwarn(
                        f"fsf found in measure {measure.measureNumber} in file {file_name}"
                    )
                    name = "f"

                new_dyn = get_dynamic_numeric(name.strip())
                if new_dyn != last_dyn:
                    (
                        beats_section,
                        dyn_grad,
                        last_dyn,
                        first_silence,
                        dyn_mean_weighted,
                    ) = calculate_dynamics(
                        dynamics,
                        beats_section,
                        dyn_mean_weighted,
                        dyn_grad,
                        last_dyn,
                        number_of_beats,
                        element,
                        beats_timesignature,
                        new_dyn,
                    )
                name = ""
                dyn = False

        beats_section += number_of_beats
        total_beats += number_of_beats
        if measure in part_data["sounding_measures"]:
            total_sounding_beats += number_of_beats

    dyn_mean_weighted += beats_section * dynamics[-1] if len(dynamics) != 0 else 0

    part_features.update(
        {
            DYNMEAN: mean(dynamics) if len(dynamics) != 0 else 0,
            DYNMEAN_WEIGHTED: float(dyn_mean_weighted / total_sounding_beats)
            if total_sounding_beats != 0
            else 0,
            DYNGRAD: float(dyn_grad / (len(dynamics) - 1))
            if (len(dynamics) - 1) > 0
            else 0,
            DYNABRUPTNESS: float(dyn_grad / total_sounding_beats)
            if total_sounding_beats != 0
            else 0,
        }
    )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    features = {}

    dyn_means = []
    dyn_means_weighted = []
    dyn_grads = []
    dyn_abruptness = []

    for part_data, part_features in zip(parts_data, parts_features):
        part = part_data[DATA_PART_ABBREVIATION]

        features[get_part_feature(part, DYNMEAN)] = part_features[DYNMEAN]
        dyn_means.append(part_features[DYNMEAN])

        features[get_part_feature(part, DYNMEAN_WEIGHTED)] = part_features[
            DYNMEAN_WEIGHTED
        ]
        dyn_means_weighted.append(part_features[DYNMEAN_WEIGHTED])

        features[get_part_feature(part, DYNGRAD)] = part_features[DYNGRAD]
        dyn_grads.append(part_features[DYNGRAD])

        features[get_part_feature(part, DYNABRUPTNESS)] = part_features[DYNABRUPTNESS]
        dyn_abruptness.append(part_features[DYNABRUPTNESS])

    dyn_means = [i for i in dyn_means if i != 0.0]
    dyn_means_weighted = [i for i in dyn_means_weighted if i != 0.0]
    dyn_grads = [i for i in dyn_grads if i != 0.0]
    dyn_abruptness = [i for i in dyn_abruptness if i != 0.0]

    features.update(
        {
            get_score_feature(DYNMEAN): mean(dyn_means) if dyn_means else 0,
            get_score_feature(DYNMEAN_WEIGHTED): mean(dyn_means_weighted)
            if dyn_means_weighted
            else 0,
            get_score_feature(DYNGRAD): mean(dyn_grads) if dyn_grads else 0,
            get_score_feature(DYNABRUPTNESS): mean(dyn_abruptness)
            if dyn_abruptness
            else 0,
        }
    )

    score_features.update(features)


def calculate_dynamics(
    dynamics,
    beats_section,
    dyn_mean_weighted,
    dyn_grad,
    last_dyn,
    number_of_beats,
    element,
    beats_timesignature,
    new_dyn,
):
    old_beat = calculate_position(number_of_beats, element, beats_timesignature)
    dyn_mean_weighted += (beats_section + old_beat) * last_dyn
    dynamics.append(new_dyn)
    beats_section, dyn_grad, last_dyn = calculate_gradient(
        beats_section, dyn_grad, last_dyn, old_beat, new_dyn
    )
    first_silence = False
    return beats_section, dyn_grad, last_dyn, first_silence, dyn_mean_weighted


def calculate_position(number_of_beats, element, beats_timesignature):
    sub_beat = element.elements[0].beat if element.isStream else element.beat
    position = _get_beat_position(beats_timesignature, number_of_beats, sub_beat)
    old_beat = position - _get_beat_position(beats_timesignature, number_of_beats, 1)
    return old_beat  # - sum([i.duration.quarterLength for i in bar_section.elements if i.classes[0] == REST]) #all silences in the measure


def calculate_gradient(beats_section, dyn_grad, last_dyn, old_beat, new_dyn):
    if (beats_section + old_beat) > 0:
        dyn_grad += abs(new_dyn - last_dyn) / (beats_section + old_beat)
    last_dyn = new_dyn
    beats_section = -old_beat  # number of beats that has old dynamic
    return beats_section, dyn_grad, last_dyn


def get_dynamic_numeric(value):
    if value in DYNAMIC_VALUES:
        return DYNAMIC_VALUES.get(value)
    else:
        pwarn(f"Dynamic value was not identified: {value}")
        return 40  # average value in case of error



from typing import List


from music21.note import Note

from numpy import mean


from musif.common._constants import VOICE_FAMILY

from musif.config import ExtractConfiguration

from musif.extract.common import _filter_parts_data

from musif.extract.constants import DATA_FAMILY, DATA_PART_ABBREVIATION

from musif.extract.features.core.handler import DATA_LYRICS, DATA_NOTES

from musif.extract.features.prefix import get_part_feature, get_score_feature

from .constants import *

from musif.extract.features.core.constants import (
    DATA_MEASURES,
    DATA_SOUNDING_MEASURES,
    NUM_NOTES,
)


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):

    notes = part_data[DATA_NOTES]

    lyrics = part_data[DATA_LYRICS]

    if part_data[DATA_FAMILY] == VOICE_FAMILY:

        syllabic_ratio = get_syllabic_ratio(notes, lyrics)

        voice_reg = get_voice_reg(notes)

        voice_presence = len(part_data[DATA_SOUNDING_MEASURES]) / len(
            part_data[DATA_MEASURES]
        )

        part_features.update(
            {
                SYLLABLES: len(lyrics),
                SYLLABIC_RATIO: syllabic_ratio,
                VOICE_REG: voice_reg,
                VOICE_PRESENCE: voice_presence,
            }
        )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):

    parts_data = _filter_parts_data(parts_data, cfg.parts_filter)

    if len(parts_data) == 0:
        return

    voice_parts_data = [
        part_data for part_data in parts_data if part_data[DATA_FAMILY] == VOICE_FAMILY
    ]

    features = {}

    if voice_parts_data:

        for part_data in voice_parts_data:

            part = part_data[DATA_PART_ABBREVIATION]

            features[get_part_feature(part, NUM_NOTES)] = len(part_data[DATA_NOTES])

            features[get_part_feature(part, SYLLABIC_RATIO)] = get_syllabic_ratio(
                part_data[DATA_NOTES], part_data[DATA_LYRICS]
            )

            features[get_part_feature(part, SYLLABLES)] = len(part_data[DATA_LYRICS])

            features[get_part_feature(part, VOICE_PRESENCE)] = len(
                part_data[DATA_SOUNDING_MEASURES]
            ) / len(part_data[DATA_MEASURES])

            features[get_part_feature(part, SYLLABIC_RATIO)] = get_syllabic_ratio(
                part_data[DATA_NOTES], part_data[DATA_LYRICS]
            )

        notes = [
            note for part_data in voice_parts_data for note in part_data[DATA_NOTES]
        ]

        lyrics = [
            lyrics
            for part_data in voice_parts_data
            for lyrics in part_data[DATA_LYRICS]
        ]

        features[get_score_feature(SYLLABLES)] = len(lyrics)

        features[get_score_feature(SYLLABIC_RATIO)] = get_syllabic_ratio(notes, lyrics)

        features[get_score_feature(VOICE_REG)] = get_voice_reg(notes)

    return score_features.update(features)


def get_syllabic_ratio(notes: List[Note], lyrics: List[str]) -> float:

    if lyrics is None or len(lyrics) == 0:

        return 0.0

    return len(notes) / len(lyrics)


def get_voice_reg(notes: List[Note]) -> float:

    if notes:

        last_note = (
            notes[-1].notes[0].pitch.midi if notes[-1].isChord else notes[-1].pitch.midi
        )

        # If we wave 2 or more notes at once, we just take the lowest one

        distances = [
            note[0].pitch.midi if note.isChord else (note.pitch.midi - last_note)
            for note in notes
        ]
        return mean(distances)

    else:

        return "NA"



class MissingFileError(Exception):
    """Exception informing that an expected file couldn't be found."""

    def __init__(self, file_path: str):
        super().__init__(f"File with path '{file_path}' couldn't be found")


class ParseFileError(Exception):
    """Exception informing that an file couldn't be parsed with the specified attributes."""

    def __init__(self, file_path: str):
        super().__init__(f"Parse error with file '{file_path}'")


class CannotResurrectObject(Exception):
    """Exception raised when a cached object is accessed on a non-cached attribute and the object doesn't know how to resurrect the non-cached object"""

    def __init__(self, obj):
        super().__init__(
            f"Cannot resurrect object {repr(obj)}, please define a `resurrect_reference` field"
        )


class SmartCacheModified(Exception):
    """Exception raised when trying to modify the reference of a `SmartCache` object"""

    def __init__(self, obj, attr):
        super().__init__(f"Cannot modify SmartCache {obj}, via attribute {attr}")


class FeatureError(RuntimeError):
    """Exception raised when computing one of the features modules"""

    pass



ALPHA = "abcdefghijklmnopqrstuvwxyz"
HARMONY_PREFIX = "Harmony_"
HARMONIC_prefix = HARMONY_PREFIX + "Harmonic"
HARMONY_AVAILABLE = HARMONY_PREFIX + "Analysis_Available"

HARMONIC_RHYTHM = HARMONIC_prefix + "Rhythm"
HARMONIC_RHYTHM_BEATS = HARMONIC_prefix + "RhythmBeats"

ADDITIONS_prefix = HARMONY_PREFIX + "Additions_"
ADDITIONS_4_6_64 = "4_6_64"
ADDITIONS_9 = "+9"
OTHERS_NO_AUG = "Others_no_+"
OTHERS_AUG = "Others_+"

KEY_PREFIX = HARMONY_PREFIX + "Key_"
KEY_GROUPING = "KeyGrouping"
KEY_MEASURES = "Measures_"
KEY_PERCENTAGE = "_PercentageMeasures"
KEY_MODULATORY = "Modulatory_"
KEY_MODCOMP = "ModComp_"

NUMERALS_prefix = HARMONY_PREFIX + "Numerals_"

CHORD_TYPES_prefix = HARMONY_PREFIX + "Chord_types_"
CHORD_prefix = HARMONY_PREFIX + "Chord_"
CHORDS_GROUPING_prefix = HARMONY_PREFIX + "Chords_Grouping_"



from typing import Tuple

from music21.key import Key
from music21.stream.base import Score


def _get_key_signature(score_key: Key) -> str:
    """
    Returns the key signature from a specific music21 Key object

    Parameters
    ----------
        score_key : Key
        Music21 Key to take the info from
    """

    if score_key.sharps:
        key_signature = (
            "b" * abs(score_key.sharps)
            if score_key.sharps < 0
            else "s" * score_key.sharps
        )
    else:
        key_signature = "n"
    return key_signature


def get_key_signature_type(key_signature: str) -> str:
    """
    Returns the key signature type ('bb) for flats, 'ss' for sharps, and 'nn' for naturals

    Parameters
    ----------
        key_signature: str
            Music21 key to take the info from
    """
    return key_signature[0]


def get_key_and_mode(score: Score) -> Tuple[Key, str, str]:
    """
    Returns abbreviated designation of keys (uppercase for major mode; lowercase for minor mode)

    Example
    ----------
    if key == 'D- major': return 'Db'

    Parameters
    ----------
        score : Score
      Music21 score to take the info from
    """

    score_key = score.analyze("key")
    mode, tonality = get_name_from_key(score_key)
    return score_key, tonality, mode


def get_name_from_key(score_key: Key) -> Tuple[str, str]:
    """
    Returns abbreviated designation of keys (uppercase for major mode; lowercase for minor mode)

    Example
    ----------
    if key == 'D- major': return 'Db'

    Parameters
    ----------
        score : Score
            Music21 score to take the info from
    """

    key_parts = score_key.name.split(" ")
    mode = key_parts[1].strip().lower()
    tonality = key_parts[0]
    tonality = tonality.lower() if mode == "minor" else tonality.capitalize()
    tonality = tonality.replace("-", "b")
    return mode, tonality


def _get_key(score: Score) -> str:
    return str(score.analyze("key"))


def _get_mode(key: str) -> str:
    if key.isupper():
        return "M"
    else:
        return "m"



from typing import List


from musif.config import ExtractConfiguration

from musif.extract.common import _filter_parts_data

from musif.extract.constants import DATA_PART_ABBREVIATION

from musif.extract.features.prefix import get_part_prefix, get_score_feature

from .utils import get_emphasised_scale_degrees_relative

from .constants import *

from musif.extract.features.core.constants import DATA_KEY, DATA_NOTES


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    if score_data:
        notes_per_degree_relative = get_emphasised_scale_degrees_relative(
            part_data[DATA_NOTES], score_data
        )
        if notes_per_degree_relative is None:
            # no harmonic data in the musescore file (or window)
            return

        all_degrees = sum(value for value in notes_per_degree_relative.values())

        for key, value in notes_per_degree_relative.items():
            part_features[DEGREE_RELATIVE_COUNT.format(key=key)] = value
            part_features[DEGREE_RELATIVE_PER.format(key=key)] = (
                value / all_degrees if all_degrees != 0 else 0
            )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):

    parts_data = _filter_parts_data(parts_data, cfg.parts_filter)

    features = {}

    if len(parts_data) == 0:
        return

    degree = score_data[DATA_KEY]

    score_notes_per_degree = {}

    for part_data in parts_data:

        notes = part_data[DATA_NOTES]
        notes_per_degree_relative = get_emphasised_scale_degrees_relative(
            part_data[DATA_NOTES], score_data
        )
        if notes_per_degree_relative is None:
            # no harmonic data in the musescore file (or window)
            return

        for degree, notes in notes_per_degree_relative.items():

            if degree not in score_notes_per_degree:

                score_notes_per_degree[degree] = 0

            score_notes_per_degree[degree] += notes

    all_score_degrees = sum(value for value in score_notes_per_degree.values())

    for degree, value in score_notes_per_degree.items():

        features[get_score_feature(DEGREE_RELATIVE_COUNT.format(key=degree))] = value

        features[get_score_feature(DEGREE_RELATIVE_PER.format(key=degree))] = (
            value / all_score_degrees if all_score_degrees != 0 else 0
        )

    for part_data, parts_features in zip(parts_data, parts_features):

        part_prefix = get_part_prefix(part_data[DATA_PART_ABBREVIATION])

        for feature in parts_features:
            if "Degree" and "relative" in feature:
                features[f"{part_prefix}{feature}"] = parts_features[feature]

    score_features.update(features)



import logging
import random
import weakref
from typing import Any, Dict, List, Optional, Tuple, Union

from deepdiff import DeepHash, deephash

from musif.common.exceptions import CannotResurrectObject, SmartCacheModified
from musif.logs import pinfo, pwarn

deephash.logger.setLevel(logging.ERROR)

CACHE_FILE_EXTENSION = ".pkl"


class ObjectReference:
    """
    This handles the calls to the reference object so that both
    `MethodCache` and `SmartModuleCache` reference the same object and when one
    resurrects it, the other sees it.
    """

    __slots__ = ("reference", "resurrect_reference", "parent", "name", "args")

    def __init__(
        self,
        reference: Any,
        resurrect_reference: Optional[Tuple],
        parent=None,
        name: str = "",
        args: Optional[Tuple] = None,
    ):
        self.reference = reference
        self.resurrect_reference = resurrect_reference
        self.parent = parent
        self.name = name
        self.args = args
        # self.deephash = DeepHash(reference)[reference]

    def _try_resurrect(self) -> None:
        if self.resurrect_reference is None:
            if self.parent is None:
                raise CannotResurrectObject(self)
            else:
                origin = self.parent
                for i in range(len(self.name)):
                    if self.name[i] == "":
                        raise CannotResurrectObject(self)
                    if isinstance(origin, ObjectReference):
                        origin = origin.get_attr(self.name[i])
                    else:
                        origin = getattr(origin, self.name[i])
                    if self.args[i] is not None:
                        origin = origin(*self.args[i])
                self.reference = origin

        else:
            func = self.resurrect_reference[0]
            args = self.resurrect_reference[1:]
            pinfo(f"Resurrecting via function call: {func}{args}")
            self.reference = func(*args)
            # self.deephash = DeepHash(self.reference)

    def get_attr(self, name: str) -> Any:
        if not hasattr(self, "reference") or self.reference is None:
            # pinfo(f"Resurrecting reference object due to call to attribute '{name}'")
            # traceback = __import__('traceback')
            # traceback.print_list(
            #     [f for f in traceback.extract_stack() if "musif" in f.filename]
            # )
            self._try_resurrect()
        return getattr(self.reference, name)

    def __getstate__(self):
        return dict(
            resurrect_reference=self.resurrect_reference,
            parent=self.parent,
            name=self.name,
            args=self.args,
        )

    def __setstate__(self, state):
        self.reference = None
        self.resurrect_reference = state["resurrect_reference"]
        self.parent = state["parent"]
        self.name = state["name"]
        self.args = state["args"]

    def __repr__(self):
        return f"ObjectReference({self.reference}, {self.name})"


class _MyNone:
    """Type used internally for representing attributes that do not exist in the
    reference object"""

    pass


class SmartModuleCache:
    """
    This class wraps any object so that its function calls can be cached in a
    dict, very similarly to how `functools.lru_cache` works.

    The object keeps a dictionary and a reference to the wrapped object.
    Each time a method, a field, or a property is called, it stores the result
    value in the dictionary.
    When the same call is performed a second time, the result value is taken
    from the dictionary.

    If the object is changed, the change is not cached. As such, only read
    operations can be executed on a ``SmartModuleCache``.

    After unpickling, the referenced object is no longer available.
    If `resurrect_reference` is not None, it will be loaded using its value. In
    such a case, `resurrect_reference` should be a tuple with a function in the
    first position and the arguments later; the function returned value must be
    the reference object.

    If the method/property/field returns an object defined in a
    one of the `target_addresses`, this class will return a wrapped version of
    that object.
    In `musif`, this is used to wrap all the music21 objects and to cache
    (most of) music21 operations.

    When a method is called, it is matched with all the arguments, similarly to
    `functools.lru_cache`, thus using the hash value of the objects.

    When pickled, this object only pickles the cache dictionary.
    The matching of the arguments in a method works on a custom hash value that
    is guaranteed to be the same when the object is pickled/unpickled. As
    consequence, as long as the method arguments are `SmartModuleCache`
    objects, they will re-use the cache after pickling/unpickling.

    Sometimes, the methods of the referenced object expect a non-cached module,
    for instance when a type checking or a hash comparison is done. For those
    situations, `SmartModuleCache` provides *special methods* that start with
    `SmartModuleCache.SPECIAL_METHODS_NAME` (which defaults to
    `"smartcache__"`). It is possible call any method with
    `smartcache__methodname`, e.g. `score.smartcache__remove`.
    When a special method is called, the first call is performed with the
    non-cached arguments, as it would happen without `SmartModuleCache`.

    NotImplemented: When `smart_pickling` is True, it pickles the dictionary
    alone, without the reference, otherwise, it pickles the reference as well.

    Note: this class implements `__repr__` in a custom way, while proxies the
    `__str__` method to the referenced object.

    Note: In future, `deepdiff.Delta` could be used to allow in-place
    operations, but it needs that:
        
    1. the object modified are pickable
    2. the reference are deep-copiable (for some reason, music21 objects)
        sometimes exceed the maximum number of recursion while deep-copying
    """

    __slots__ = "cache"
    SPECIAL_METHODS_NAME = "smartcache__"

    def __init__(
        self,
        reference: Any = None,
        resurrect_reference: Optional[Tuple] = None,
        parent: Optional[ObjectReference] = None,
        name: Tuple[str] = ("",),
        args: Tuple[Optional[Tuple]] = (None,),
        target_addresses: List[str] = ["music21"],
        check_reference_changes: bool = False,
    ):

        cache: Dict[
            str,
            Union[Tuple, str, int, SmartModuleCache, MethodCache, Any],
        ] = {
            "_hash": random.randint(10**10, 10**15),
            "_target_addresses": target_addresses,
            "_reference": ObjectReference(
                reference, resurrect_reference, parent, name, args
            ),
            "_check_reference_changes": check_reference_changes,
            "_type": type(reference),
        }
        object.__setattr__(self, "cache", cache)

    def smartforcecache(self, name, *args, **kwargs):
        """
        This method can be used to forcedly cache some attribute of the reference
        object without using it. `name` is the name of the attribute. If it is a method,
        `*args` and `**kwargs`.

        Return `(True, None)` if object is cached, `(False, Exception)` otherwise
        """
        try:
            attr = self.__getattr__(name)
            assert self.cache["name"] is not None, "Object retrieved but not cached!"

            if isinstance(attr, MethodCache):
                attr.__forcecache(*args, **kwargs)
        except Exception as e:
            return False, e
        return True, None

    def __repr__(self):
        _reference = self.cache["_reference"]
        _addresses = self.cache["_target_addresses"]
        _class = self.__class__.__qualname__
        return f"{_class}({_reference}, {_addresses})"

    def __str__(self):
        s = self.cache.get("__str__")
        if s is None:
            s = str(self.cache["_reference"].reference)
            self.cache["__str__"] = s
        return s

    @property
    def target_addresses(self):
        return self.cache["_target_addresses"]

    @target_addresses.setter
    def target_addresses(self, value):
        self.cache["_target_addresses"] = value

    def __hash__(self):
        return self.cache["_hash"]

    def _wmo(self, obj, name=("",), args=(None,)):
        from .utils import wrap_module_objects
        return wrap_module_objects(
            obj,
            target_addresses=self.cache["_target_addresses"],
            resurrect_reference=None,
            parent=self.cache["_reference"],
            name=name,
            args=args,
        )

    def __getattr__(self, name: str) -> Any:
        """
        The real key of this class is this method!
        """
        attr = self.cache.get(name)
        if attr is None:
            # attr not in cache
            attr = self._get_new_attr(name)
            if attr is None:
                attr = _MyNone
            self.cache[name] = attr
        if attr is _MyNone:
            return None
        else:
            return attr

    def __delattr__(self, name):
        del self.cache[name]

    def __setattr__(self, name, value):
        pwarn(
            "You are using a music21 object in writing mode! consider change your code so that SmartModuleCache works correctly. See the stack-trace:"
        )
        # traceback = __import__('traceback')
        # traceback.print_list(
        #     [f for f in traceback.extract_stack() if "musif" in f.filename]
        # )
        if self.cache["_reference"].reference is not None:
            self.cache["_reference"].get_attr("__setattr__")(name, value)

    def _get_new_attr(self, name: str) -> Any:
        if name.startswith(self.SPECIAL_METHODS_NAME):
            special_name = name
            name = name[len(self.SPECIAL_METHODS_NAME) :]
        else:
            special_name = None
        attr = self.cache["_reference"].get_attr(name)
        if callable(attr):
            # use the method cacher
            attr = MethodCache(
                self.cache["_reference"],
                name,
                target_addresses=self.cache["_target_addresses"],
                special_method=special_name is not None,
                check_reference_changes=self.cache["_check_reference_changes"],
            )
        else:
            # cache the value and returns it
            attr = self._wmo(attr, (name,))

        return attr

    def __setstate__(self, state):
        object.__setattr__(self, "cache", state)

    def __getstate__(self):
        return self.cache

    # caching other special methods
    def __list__(self, idx=None):
        it = self.cache.get("__list__")
        if it is None:
            it = self.cache["_reference"].get_attr("__iter__")()
            it = [
                self._wmo(
                    x,
                    ("__getitem__",),
                    ((i,),),
                )
                for i, x in enumerate(it)
            ]
            self.cache["__list__"] = it
        if idx is not None:
            return it[idx]
        else:
            return it

    def __iter__(self):
        return iter(self.__list__())

    def __len__(self):
        L = self.cache.get("__len__")
        if L is None:
            L = self.cache["_reference"].get_attr("__len__")()
            self.cache["__len__"] = L
        return L

    def __getitem__(self, k):
        v = self.cache.get(f"__item_{k}")
        if v is None:
            v = self.cache["_reference"].get_attr("__getitem__")(k)
            v = self._wmo(v, ("__getitem__",), ((k,),))
            self.cache[f"__item_{k}"] = v
        return v

    def __setitem__(self, k, v):
        pwarn(
            "You are using a music21 object in writing mode! consider change your code so that SmartModuleCache works correctly. See the stack-trace:"
        )
        # traceback = __import__('traceback')
        # traceback.print_list(
        #     [f for f in traceback.extract_stack() if "musif" in f.filename]
        # )
        try:
            v = self.cache["_reference"].get_attr("__getitem__")(k)
        except KeyError:
            pass
        v = self._wmo(v, ("__setitem__",), ((k, v),))
        self.cache[f"__item_{k}"] = v
        self.cache["_reference"].get_attr("__setattr__")(k, v)

    def __bool__(self):
        b = self.cache.get("__bool__")
        if b is None:
            try:
                b = self.cache["_reference"].get_attr("__bool__")()
            except AttributeError:
                b = self.cache["_reference"].reference is not None
            finally:
                self.cache["__bool__"] = b
        return b

    def ischanged(self):
        """
        Returns if the reference object is changed using value comparison of
        its attributes recursively.
        """
        newhash = DeepHash(self.reference)[self.reference]
        if newhash != self.deephash:
            return True
        else:
            return False


class CallableArguments:
    """
    This class represents a set of ordered arguments.
    The hash is the concatenation of the argument hashes.
    If the arguments are SmartModuleCache objects, this object uses their cache (which
    persists to disk), otherwise it uses `deepdiff.DeepHash` to compute a hash based on
    the content value (which should persists across pickling).

    Note that `DeepHash` may be much slower than the builtin `hash()` function,
    so for large and complex objects like `music21.Score`, just use a
    `SmartModuleCache` wrapper.
    """

    def __init__(self, *args, **kwargs):
        self.args = []
        self.kwargs = {}

        # removing weakreferences (cannot be pickled)
        for arg in args:
            if isinstance(arg, weakref.ReferenceType):
                __import__('ipdb').set_trace()
                self.args.append(arg())
            else:
                self.args.append(arg)

        for k, v in kwargs.items():
            if isinstance(k, weakref.ReferenceType):
                __import__('ipdb').set_trace()
                k = k()
            if isinstance(v, weakref.ReferenceType):
                __import__('ipdb').set_trace()
                v = v()
            kwargs[k] = v

        h = ""

        # hashing SmartModuleCache
        for arg in self.args:
            if type(arg) is SmartModuleCache:
                h += str(hash(arg))

        kwargs_keys = sorted(self.kwargs.keys())
        for k in kwargs_keys:
            v = self.kwargs[k]
            if type(v) is SmartModuleCache:
                h += str(hash(v))

        # hashing other object types by value
        all_args = (self.args, self.kwargs)
        h += DeepHash(all_args, exclude_types=[SmartModuleCache])[all_args]

        self._hash = int(h, 16)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if type(other) is CallableArguments:
            return other._hash == self._hash
        else:
            return False

    def __repr__(self):
        ret = "CallableArguments("
        for a in self.args:
            ret += f"{repr(self.args)}, "
        for k, v in self.kwargs.items():
            ret += f"{repr(k)}={repr(v)}, "
        ret += ")"
        return ret


class MethodCache:
    """
    A simple wrapper that checks if the arguments are in the
    cache, otherwise runs the method and caches the arguments
    """

    __slots__ = (
        "reference",
        "name",
        "cache",
        "target_addresses",
        "special_method",
        "check_reference_changes",
    )

    def __init__(
        self,
        reference: ObjectReference,
        attr_name: str,
        target_addresses: List[str] = ["music21"],
        special_method: bool = False,
        check_reference_changes: bool = False,
    ):
        self.reference = reference
        self.name = attr_name
        self.cache = dict()
        self.target_addresses = target_addresses
        self.special_method = special_method
        self.check_reference_changes = check_reference_changes

    def _wmo(self, obj, args=None):
        from .utils import wrap_module_objects
        return wrap_module_objects(
            obj,
            target_addresses=self.target_addresses,
            resurrect_reference=None,
            parent=self.reference,
            name=(self.name,),
            args=(args,),
        )

    def __call__(self, *args, **kwargs):
        args = list(map(self._wmo, args))
        kwargs = {k: self._wmo(v) for k, v in kwargs.items()}
        call_args = CallableArguments(*args, **kwargs)
        cached_res = self.cache.get(call_args)
        if cached_res is not None:
            if cached_res is _MyNone:
                return None
            return cached_res
        else:
            attr = self.reference.get_attr(self.name)
            if self.special_method:
                args = [arg.cache["_reference"].reference for arg in args]
                kwargs = {k: v.cache["_reference"].reference for k, v in kwargs.items()}
            res = attr(*args, **kwargs)
            if res is None:
                # caching _MyNone and returning None
                self.cache[call_args] = _MyNone
            else:
                res = self._wmo(res, args=(*args, *kwargs.values()))
                self.cache[call_args] = res
            if self.check_reference_changes and self.reference.ischanged():
                pwarn(str(SmartCacheModified(self, self.name)))
            return res

    def smartforcecache(self, *args, **kwargs):
        self.__call__(*args, **kwargs)
        assert (
            self.cache.get(CallableArguments(*args, **kwargs)) is not None
        ), "Method retrieved but arguments not cached!"



import re
from logging.config import dictConfig
from typing import List

import pandas as pd
from pandas import DataFrame

from musif.config import (
    ENDSWITH,
    INSTRUMENTS_TO_DELETE,
    INSTRUMENTS_TO_KEEP,
    STARTSWITH,
)
from musif.extract.basic_modules.scoring.constants import (
    FAMILY_INSTRUMENTATION,
    FAMILY_SCORING,
)
from musif.extract.features.harmony.constants import (
    KEY_MODULATORY,
    KEY_PERCENTAGE,
    KEY_PREFIX,
    CHORD_prefix,
)
from musif.extract.features.melody.constants import TRIMMED_INTERVALLIC_MEAN
from musif.extract.features.prefix import get_part_prefix
from musif.extract.features.scale.constants import DEGREE_PREFIX
from musif.logs import pinfo

from .constants import voices_list_prefixes


def replace_nans(df):
    for col in df.columns:
        if (
            "Interval" in col
            or col.startswith("Key_")
            or col.startswith((CHORD_prefix, "Chords_", "Additions_", "Numerals_"))
            or col.endswith(("_DottedRhythm", "_DoubleDottedRhythm"))
            or ("_Degree" and TRIMMED_INTERVALLIC_MEAN and "_Dyn") in col
        ):
            df[col] = df[col].fillna("NA")


def join_part_degrees(
    total_degrees: List[str], part: str, df: DataFrame, sufix: str = ""
) -> None:
    part_degrees = [i for i in total_degrees if part in i]

    aug = [i for i in part_degrees if "#" in i]
    desc = [i for i in part_degrees if "b" in i and "bb" not in i]
    d_desc = [i for i in part_degrees if "bb" in i]
    d_asc = [i for i in part_degrees if "x" in i]

    pattern = "^" + part + "Degree" + "[0-9].*"
    degree_nat = [x for x in part_degrees if re.search(pattern, x)]
    degree_nonat = [i for i in part_degrees if i not in degree_nat]

    # TODO: strange names here, should be like aug, dim, etc.
    df[part + DEGREE_PREFIX + "_Asc" + sufix] = df[aug].sum(axis=1)
    df[part + DEGREE_PREFIX + "_Desc" + sufix] = df[desc].sum(axis=1)
    df[part + DEGREE_PREFIX + "_Dasc" + sufix] = df[d_asc].sum(axis=1)
    df[part + DEGREE_PREFIX + "_Ddesc" + sufix] = df[d_desc].sum(axis=1)
    df[part + DEGREE_PREFIX + "_Nat" + sufix] = df[degree_nat].sum(axis=1)
    df[part + DEGREE_PREFIX + "_Nonat" + sufix] = df[degree_nonat].sum(axis=1)


def log_errors_and_shape(
    composer_counter: list, novoices_counter: list, df: DataFrame
) -> None:
    pinfo(f"\nTotal files skipped by composer: {len(composer_counter)}")
    pinfo(str(composer_counter))
    pinfo(f"\nTotal files skipped by no-voices: { len(novoices_counter)}")
    pinfo(str(novoices_counter))
    # pinfo(f"\nTotal files skipped by duetos/trietos: {len(duetos_counter)}")
    # pinfo(str(duetos_counter))
    pinfo(f"\nFinal shape of the DataFrame: {df.shape[0]} rows, {df.shape[1]} features")


def _delete_columns(data: DataFrame, config: dictConfig) -> None:
    # pinfo("\nDeleting not useful columns...")
    to_delete = []
    instruments_to_keep = [get_part_prefix(i) for i in config[INSTRUMENTS_TO_KEEP]]
    for inst in config[INSTRUMENTS_TO_DELETE]:
        # for i in data.columns:
        #     if "Part" + inst + "_" in i
        part_prefix = "Part" + inst  # + "_"
        for col in data.columns:
            if part_prefix in col and all(
                inst not in col for inst in instruments_to_keep
            ):
                pass
                to_delete.append(col)
            else:
                pass
        # to_delete += [i for i in data.columns if part_prefix in i and instrument not in i for instrument in instruments_to_keep]

    to_delete += [i for i in data.columns if i.endswith(tuple(config[ENDSWITH]))]
    to_delete += [i for i in data.columns if i.startswith(tuple(config[STARTSWITH]))]
    to_delete += [
        col
        for col in data.columns
        if any(substring in col for substring in config["columns_contain"])
    ]
    to_delete += [
        col
        for col in data.columns
        if any(string == col for string in config["columns_match"])
    ]
    to_delete += [i for i in data.columns if i.startswith("Sound") and "Voice" not in i]

    to_delete += [FAMILY_INSTRUMENTATION, FAMILY_SCORING]

    # Remove empty voices
    to_delete += [
        col
        for col in data.columns
        if col.startswith(tuple(voices_list_prefixes))
        and all(data[col].isnull().values)
    ]

    # removing columns containing nans
    th = config["max_nan_columns"] or 0.0
    idx = data.isna().sum(axis=0) / data.shape[0] > th
    to_delete += data.columns[idx].to_list()

    data.drop(columns=to_delete, inplace=True, errors="ignore")


def join_keys(df: DataFrame) -> None:
    key_SD = [
        i
        for i in [
            KEY_PREFIX + "IV" + KEY_PERCENTAGE,
            KEY_PREFIX + "II" + KEY_PERCENTAGE,
            KEY_PREFIX + "VI" + KEY_PERCENTAGE,
        ]
        if i in df
    ]
    key_sd = [
        i
        for i in [
            KEY_PREFIX + "iv" + KEY_PERCENTAGE,
            KEY_PREFIX + "ii" + KEY_PERCENTAGE,
        ]
        if i in df
    ]
    key_tonic = [
        i
        for i in [KEY_PREFIX + "I" + KEY_PERCENTAGE, KEY_PREFIX + "i" + KEY_PERCENTAGE]
        if i in df
    ]
    key_rel = [
        i
        for i in [
            KEY_PREFIX + "III" + KEY_PERCENTAGE,
            KEY_PREFIX + "vi" + KEY_PERCENTAGE,
        ]
        if i in df
    ]

    total_key = key_rel + key_tonic + key_sd + key_SD
    others_key = [
        i
        for i in df.columns
        if KEY_PREFIX in i and i not in total_key and KEY_MODULATORY not in i
    ]

    df[KEY_PREFIX + "SD" + KEY_PERCENTAGE] = df[key_SD].sum(axis=1)
    df[KEY_PREFIX + "sd" + KEY_PERCENTAGE] = df[key_sd].sum(axis=1)
    df[KEY_PREFIX + "SubD" + KEY_PERCENTAGE] = (
        df[KEY_PREFIX + "sd" + KEY_PERCENTAGE] + df[KEY_PREFIX + "SD" + KEY_PERCENTAGE]
    )
    df[KEY_PREFIX + "T" + KEY_PERCENTAGE] = df[key_tonic].sum(axis=1)
    df[KEY_PREFIX + "rel" + KEY_PERCENTAGE] = df[key_rel].sum(axis=1)
    df[KEY_PREFIX + "Other" + KEY_PERCENTAGE] = df[others_key].sum(axis=1)
    # df.drop(total_key + others_key, axis = 1, inplace=True)


def join_keys_modulatory(df: DataFrame):
    key_SD = [
        i
        for i in [
            KEY_PREFIX + KEY_MODULATORY + "IV",
            KEY_PREFIX + KEY_MODULATORY + "II",
            KEY_PREFIX + KEY_MODULATORY + "VI",
        ]
        if i in df
    ]
    key_sd = [
        i
        for i in [
            KEY_PREFIX + KEY_MODULATORY + "iv",
            KEY_PREFIX + KEY_MODULATORY + "ii",
        ]
        if i in df
    ]
    key_tonic = [
        i
        for i in [KEY_PREFIX + KEY_MODULATORY + "I", KEY_PREFIX + KEY_MODULATORY + "i"]
        if i in df
    ]
    key_rel = [
        i
        for i in [
            KEY_PREFIX + KEY_MODULATORY + "III",
            KEY_PREFIX + KEY_MODULATORY + "vi",
        ]
        if i in df
    ]

    total_key_mod = key_rel + key_tonic + key_sd + key_SD
    others_key_mod = [
        i
        for i in df.columns
        if KEY_PREFIX + KEY_MODULATORY in i and i not in total_key_mod
    ]

    df[KEY_PREFIX + KEY_MODULATORY + "SD"] = df[key_SD].sum(axis=1)
    df[KEY_PREFIX + KEY_MODULATORY + "sd"] = df[key_sd].sum(axis=1)
    df[KEY_PREFIX + KEY_MODULATORY + "SubD"] = (
        df[KEY_PREFIX + KEY_MODULATORY + "sd"] + df[KEY_PREFIX + KEY_MODULATORY + "SD"]
    )
    df[KEY_PREFIX + KEY_MODULATORY + "T"] = df[key_tonic].sum(axis=1)
    df[KEY_PREFIX + KEY_MODULATORY + "rel"] = df[key_rel].sum(axis=1)
    df[KEY_PREFIX + KEY_MODULATORY + "Other"] = df[others_key_mod].sum(axis=1)


def merge_dataframes(name: str, dest_path: str) -> None:
    csv = ".csv"
    name1 = name + "_1" + csv
    name2 = name + "_2" + csv

    df1 = pd.read_csv(name1)
    df2 = pd.read_csv(name2)
    total_dataframe = pd.concat((df1, df2), axis=0)
    total_dataframe.to_csv(dest_path, index=False)



from statistics import mean
from typing import List, Tuple

from musif.common.sort import sort_dict
from musif.config import ExtractConfiguration
from musif.extract.common import _filter_parts_data, _part_matches_filter
from musif.extract.constants import (
    DATA_FAMILY_ABBREVIATION,
    DATA_PART_ABBREVIATION,
    DATA_SOUND_ABBREVIATION,
)
from musif.extract.features.core.handler import (
    DATA_MEASURES,
    DATA_NOTES,
    DATA_SOUNDING_MEASURES,
)
from musif.extract.features.prefix import (
    get_family_feature,
    get_family_prefix,
    get_part_feature,
    get_part_prefix,
    get_score_feature,
    get_score_prefix,
    get_sound_feature,
    get_sound_prefix,
)
from musif.extract.basic_modules.scoring.constants import NUMBER_OF_FILTERED_PARTS
from musif.extract.features.tempo.constants import (
    NUMBER_OF_BEATS,
    TIME_SIGNATURE,
    TIME_SIGNATURES,
)
from musif.extract.utils import (
    _calculate_total_number_of_beats,
)
from musif.logs import lerr
from musif.musicxml import Measure, Note, Part
from musif.musicxml.tempo import get_number_of_beats
from .constants import *
from musif.extract.features.core.constants import (
    NUM_MEASURES,
    NUM_NOTES,
    NUM_SOUNDING_MEASURES,
)
from musif.cache import isinstance


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    if not _part_matches_filter(part_data[DATA_PART_ABBREVIATION], cfg.parts_filter):
        return {}
    notes = part_data[DATA_NOTES]
    sounding_measures = part_data[DATA_SOUNDING_MEASURES]
    # measures = part_data[DATA_MEASURES]
    # time_signature = score_data[TIME_SIGNATURE].split(",")
    time_signatures = part_data[TIME_SIGNATURES]
    # sounding_measures = range(
    #     0, len(sounding_measures)
    # )  
    ## TODO: cuando haya repeticiones, revisar esto. Lo hice por un error en la numeracion cuando hay 70x1 (celdillas)
    sounding_time_signatures = [time_signatures[i] for i in sounding_measures]

    part_features.update(
        {
            SOUNDING_DENSITY: len(notes)
            / _calculate_total_number_of_beats(sounding_time_signatures)
            if len(sounding_time_signatures) > 0
            else 0,
            DENSITY: len(notes)
            / _calculate_total_number_of_beats(time_signatures)
            if len(time_signatures) > 0 else 0,
        }
    )


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):

    parts_data = _filter_parts_data(parts_data, cfg.parts_filter)
    if len(parts_features) == 0:
        return

    time_signature = score_data[TIME_SIGNATURE].split(",")
    number_of_beats = get_number_of_beats(time_signature[0])

    features = {}
    for i, part_data in enumerate(parts_data):
        part_abbreviation = part_data[DATA_PART_ABBREVIATION.lower()]
        feature_name = get_part_feature(part_abbreviation, SOUNDING_DENSITY)

        if feature_name in features:
            features[feature_name] = mean(
                [features[feature_name], parts_features[i][SOUNDING_DENSITY]]
            )
        else:
            features[feature_name] = parts_features[i][SOUNDING_DENSITY]

        feature_name = get_part_feature(part_abbreviation, DENSITY)
        if feature_name in features:
            features[feature_name] = mean(
                [features[feature_name], parts_features[i][DENSITY]]
            )
        else:
            features[feature_name] = parts_features[i][DENSITY]

    sound_abbreviations = {
        part_data[DATA_SOUND_ABBREVIATION] for part_data in parts_data
    }
    for sound in sound_abbreviations:
        num_parts = score_data[get_sound_feature(sound, NUMBER_OF_FILTERED_PARTS)]
        num_measures = score_features[NUM_MEASURES] * num_parts
        num_sounding_measures = score_features[
            get_sound_feature(sound, NUM_SOUNDING_MEASURES)
        ]
        num_notes = score_features[get_sound_feature(sound, NUM_NOTES)]
        features[get_sound_feature(sound, SOUNDING_DENSITY)] = (
            num_notes / number_of_beats / num_sounding_measures
            if num_sounding_measures != 0
            else 0
        )
        features[get_sound_feature(sound, DENSITY)] = (
            num_notes / number_of_beats / num_measures if num_measures != 0 else 0
        )

    family_abbreviations = {
        part_data[DATA_FAMILY_ABBREVIATION] for part_data in parts_data
    }
    for family in family_abbreviations:
        num_parts = score_data[get_family_feature(family, NUMBER_OF_FILTERED_PARTS)]
        num_measures = score_features[NUM_MEASURES] * num_parts
        num_sounding_measures = score_features[
            get_family_feature(family, NUM_SOUNDING_MEASURES)
        ]
        num_notes = score_features[get_family_feature(family, NUM_NOTES)]
        features[get_family_feature(family, SOUNDING_DENSITY)] = (
            num_notes / number_of_beats / num_sounding_measures
            if num_sounding_measures != 0
            else 0
        )
        features[get_family_feature(family, DENSITY)] = (
            num_notes / number_of_beats / num_measures if num_measures != 0 else 0
        )

    num_parts = score_data[get_score_feature(NUMBER_OF_FILTERED_PARTS)]
    num_measures = score_features[NUM_MEASURES] * num_parts
    num_sounding_measures = score_features[get_score_feature(NUM_SOUNDING_MEASURES)]
    num_notes = score_features[get_score_feature(NUM_NOTES)]
    features[get_score_feature(SOUNDING_DENSITY)] = (
        num_notes / number_of_beats / num_sounding_measures
        if num_sounding_measures != 0
        else 0
    )
    features[get_score_feature(DENSITY)] = (
        num_notes / number_of_beats / num_measures if num_measures != 0 else 0
    )

    score_features.update(features)


def get_notes_and_measures(
    part: Part,
) -> Tuple[List[Note], List[Measure], List[Measure]]:
    notes = []
    measures = list(part.measures(0, None))
    sounding_measures = []
    for measure in measures:
        if len(measure.notes) > 0:
            sounding_measures.append(measure)
        for note in measure.notes:
            set_ties(note, notes)
    return notes, sounding_measures, measures


def set_ties(subject, my_notes_list):
    """
    This function converts tied notes into a unique note
    """
    if subject.tie is None:
        my_notes_list.append(subject)
        return
    if subject.tie.type != "stop" and subject.tie.type != "continue":
        my_notes_list.append(subject)
        return
    if isinstance(my_notes_list[-1], Note):
        # sum tied notes' length
        my_notes_list[-1].duration.quarterLength += subject.duration.quarterLength
        return
    back_counter = -1
    while isinstance(my_notes_list[back_counter], tuple):
        back_counter -= -1
    else:
        my_notes_list[
            back_counter
        ].duration.quarterLength += (
            subject.duration.quarterLength
        )  # sum tied notes' length across measures


def calculate_densities(notes_list, measures_list, names_list, cfg: ExtractConfiguration):
    density_list = []
    try:
        for i, part in enumerate(names_list):
            density = round(notes_list[i] / measures_list[i], 3)
            density_list.append({f"{names_list[i]}": density})

        density_dict = dict((key, d[key]) for d in density_list for key in d)
        density_sorting = cfg.scoring_order
        density_dict = sort_dict(density_dict, density_sorting)
        return density_dict
    except:
        lerr("Densities problem found: ", exc_info=True)
        return {}



from statistics import mean

from typing import List, Optional


from musif.extract.constants import DATA_PART_ABBREVIATION, VOICES_LIST


def _filter_parts_data(
    parts_data: List[dict], parts_filter: Optional[List[str]]
) -> List[dict]:

    if parts_filter is None or len(parts_filter) == 0:
        return parts_data

    if "voice" in parts_filter:
        parts_filter.remove("voice")
        parts_filter.extend(VOICES_LIST)

    parts_filter_set = set(parts_filter)

    return [
        part_data
        for part_data in parts_data
        if part_data[DATA_PART_ABBREVIATION] in parts_filter_set
    ]


def _part_matches_filter(
    part_abbreviation: str, parts_filter: Optional[List[str]]
) -> bool:

    if parts_filter is None or len(parts_filter) == 0:
        return True
    parts_filter_set = set(parts_filter)
    return part_abbreviation in parts_filter_set


def _mix_data_with_precedent_data(prev_features: dict, new_features: dict) -> None:
    for key in new_features.keys():
        if "max" in key.lower() or "highest" in key.lower():
            prev_features[key] = max(prev_features[key], new_features[key])
            continue

        elif "min" in key.lower() or "lowest" in key.lower():
            prev_features[key] = min(prev_features[key], new_features[key])
            continue

        if prev_features[key] is None:
            prev_features[key] = new_features[key]
        elif new_features[key] is None:
            continue
        elif type(new_features[key]) != str:
             prev_features[key] = mean([prev_features[key], new_features[key]])



TEMPO = "Tempo"
NUMERIC_TEMPO = "NumericTempo"
TIME_SIGNATURE = "TimeSignature"
TIME_SIGNATURES = "AllTimeSignatures"
TS_MEASURES = "TimeSignatureMeasures"
TIME_SIGNATURE_GROUPED = "TimeSignatureGrouped"
TEMPO_GROUPED_1 = "TempoGrouped1"
TEMPO_GROUPED_2 = "TempoGrouped2"
NUMBER_OF_BEATS = "NumberOfBeats"



from typing import List

import music21 as m21
from music21.features import jSymbolic, native
from music21.features.base import extractorsById

from musif import cache
from musif.config import ExtractConfiguration
from musif.extract.constants import DATA_SCORE
from .constants import ERRORED_NAMES


def allFeaturesAsList(streamInput, includeJSymbolic=True):
    """
    only a little change around m21.features.base.allFeaturesAsList: no Parallel
    processing
    """

    ds = m21.features.base.DataSet(classLabel="")
    ds.runParallel = False  # this is the only difference with the m21 original code
    all_features = list(native.featureExtractors)
    if includeJSymbolic:
        all_features += list(jSymbolic.featureExtractors)
    f = [feature for feature in all_features if feature.id not in
         ERRORED_NAMES]
    # f = list(native.featureExtractors)
    ds.addFeatureExtractors(f)
    ds.addData(streamInput)
    ds.process()
    allData = ds.getFeaturesAsList(
        includeClassLabel=False, includeId=False, concatenateLists=False
    )

    return allData


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    score = score_data[DATA_SCORE]
    # override the isinstance and hasattr definitions for the caching system
    m21.features.base.isinstance = cache.isinstance
    m21.features.base.hasattr = cache.hasattr
    # avoid extracting jsymbolic features twice
    includeJSymbolic = 'jsymbolic' in cfg.features
    features = allFeaturesAsList(score, includeJSymbolic=includeJSymbolic)
    if includeJSymbolic:
        all_columns = [x.id for x in extractorsById("all")]
    else:
        all_columns = [x.id for x in extractorsById("native")]
    columns = [c for c in all_columns if c not in ERRORED_NAMES]
    score_features.update(
        {
            'm21_' + columns[outer] + f"_{i}": f
            for outer in range(len(columns))
            for i, f in enumerate(features[outer])
        }
    )


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    pass



import sys
from typing import Optional
from pathlib import Path

import musif.musicxml.constants as musicxml_c
from musif.config import ExtractConfiguration, PostProcessConfiguration
from musif.extract.extract import FeaturesExtractor
from musif.logs import perr, pinfo
from musif.process.processor import DataProcessor


def main(
    *paths,
    output_path: str = "musif_features.csv",
    source_dir: str = None,
    extension: str = ".xml",
    njobs: int = -1,
    cache_dir: str = "musif_cache",
    ignore_errors: bool = True,
    yaml: str = None,
    tweaks: dict = {},
    harmony: Optional[str] = None,
):
    """
    Python tool for extracting features from music score files.

    This tool uses `music21` to load files, so any file format supported by `music21`
    also works, e.g. MIDI, MusicXML, Kern, ABC files. It uses cache, parallel
    processing, and ignore errors by default. See the options to disable them.

    This tool uses a default configuration that should work well in most cases. See
    [paper] for more benchmarks. By default, it extracts all features except the ones
    that require harmonic annotations. You can use `-y/--yaml` and `-c/--config` for
    more tweaks.

    Examples of usage:
        musif dataset/**/*.mid
            -> process all the midi files in `dataset`, searching files recursively in
            all the sub-directories
        musif *.xml --cache_dir=None
            -> process all xml files in this directory without using cache
        musif --source_dir=dataset --extension=.krn
            -> process all kern files in the directory `dataset`, searching files
            recursively in all the sub-drectories
        musif -- -h
            -> shows this help

    Installation:
        If you have musif installed in your environment, you can access this tool using
        `python -m musif`.

        Alternatively, you can install musif system-wide with `pipx install musif`
        (see pipx documentation for instructions: https://pypa.github.io/pipx/)

    Args:
        paths : one or more paths to file; if provided, the extraction is limited to
            them; these paths can be absolute or relative to the current
            directory; all the paths should contain a common parent part;
            incompatible with `--source_dir`
        output_path : output file; extension is added or changed to 'csv'
        extension : extension, including the dot, e.g. '.mid', '.krn', '.mxl'; only
            has effect if `source_dir` is used, otherwise it's
            inferred from the file names; if one of ['.mxl', '.xml',
            '.musicxml'] is passed, all the other used as well; you can
            set it to a list or to a tuple to include multiple
            extensions;
        source_dir : relative path to the directory; searched recursively;
            incompatible with other files provided
        njobs : number of jobs used, according to joblib: -1 means "all the
            available virtual cores"; 1 means no parallel processing
        cache_dir : directory where cache files are saved; set to 'None' to disable
        ignore_errors : True or False; if False, blocks when a file cannot be
            processed, if True, cprints a warning and continue
        yaml : path to a configuration file that will be used for both extraction and
            post-processing; command line options have the precedence on this yaml file
        tweaks : Further flags can be used to change musif's configuration
            (see the docs for possible options); for this, you should
            pass them as a dictionary, e.g. `musif -t
            '{musescore_dir: "mscore_data"}'`
        harmony : extract harmonic features using musescore files from this directory
    """

    if source_dir is not None and len(paths) > 0:
        perr("Please, provide only one option between `source_dir` and file paths")
        sys.exit(1)

    if source_dir is None and len(paths) == 0:
        perr("Please, provide at least one option between `source_dir` and file paths")
        sys.exit(2)

    if source_dir is None:
        # look for the common parent part into `limit_to_files`
        first_file = Path(paths[0])
        source_dir = first_file.parent

        _source_dir = str(source_dir)
        extension = first_file.suffix

        for file in paths[1:]:
            file_ = Path(file)
            if file_.suffix != extension:
                perr(
                    f"Please provide files with only one extension: first file has extension {extension}, but {file_} has extension {file_.suffix}"
                )
            while not file.startswith(_source_dir):
                source_dir = source_dir.parent
                _source_dir = str(source_dir)
        pinfo(f"Detected parent directory: {_source_dir}")
        pinfo(f"Detected extension: {extension}")
    else:
        paths = None
        if extension in [".xml", ".mxl", ".musicxml"]:
            extension = [".xml", ".mxl", ".musicxml"]

    config = ExtractConfiguration(
        yaml,
        xml_dir=source_dir,
        cache_dir=cache_dir,
        parallel=njobs,
        musescore_dir=harmony,
        ignore_errors=ignore_errors,
        **tweaks,
    )
    if config.features == ["core"]:
        config.features = [
            "core",
            "ambitus",
            "melody",
            "tempo",
            "density",
            "texture",
            "lyrics",
            "scale",
            "key",
            "dynamics",
            "rhythm",
            "music21",
        ]
    if harmony is not None:
        config.features += ["harmony", "scale_relative"]
    if len(config.basic_modules) == 0:
        config.basic_modules = ["scoring"]
    musicxml_c.MUSIC21_FILE_EXTENSIONS = extension
    raw_df = FeaturesExtractor(config, limit_files=paths).extract()

    output_path = Path(output_path).with_suffix(".csv")
    # raw_df.to_csv(output_path.with_suffix(".raw.csv"), index=False)

    config = PostProcessConfiguration(yaml, **tweaks)
    if len(config.columns_contain) == 0:
        config.columns_contain = [
            "_Count",
            "_SmallestInterval",
            "_NumberOfFilteredParts",
        ]
    if len(config.replace_nans) == 0:
        config.replace_nans = ["Interval", "Degree", "Harmony"]
    if config.max_nan_rows is None:
        # if the columns without nans would produce a table with columns/row ratio <
        # 0.25, we remove the rows that are outliers of the 0.99 quantile
        raw_df_na = raw_df.isna()
        num_columns_without_na = (~raw_df_na.any(axis=0)).sum()
        if num_columns_without_na / raw_df.shape[0] < 0.1:
            nans = raw_df_na.sum(axis=1)
            config.max_nan_rows = 1 / 0.99 * nans.quantile(0.99) / raw_df.shape[1]
        else:
            config.max_nan_rows = 1.0
    if config.max_nan_columns is None:
        config.max_nan_columns = 0.0
    processed_df = DataProcessor(raw_df, config).process().data
    # replace empty strings with "NA" only in string columns
    string_cols = processed_df.select_dtypes(include=["object", "string"]).columns
    processed_df[string_cols] = processed_df[string_cols].replace("", "-")
    processed_df.to_csv(output_path, index=False)


if __name__ == "__main__":
    import fire

    fire.Fire(main, name="musif")



from typing import List
from musif.config import ExtractConfiguration

import contextlib
import os
import tempfile
import subprocess
import pandas as pd

import musif.extract.constants as C
from musif.logs import pwarn
from musif.extract.features.jsymbolic.utils import get_java_path, _jsymbolic_path

JSYMBOLIC_JAR = str(_jsymbolic_path())
JAVA_PATH = get_java_path()


def get_tmpdir():
    if os.path.exists("/dev/shm"):
        return tempfile.TemporaryDirectory(dir="/dev/shm")
    else:
        return tempfile.TemporaryDirectory()


def update_score_objects(
    score_data: dict,
    parts_data: List[dict],
    cfg: ExtractConfiguration,
    parts_features: List[dict],
    score_features: dict,
):
    # 1. create a temporary directory (if Linux, force RAM usig /dev/shm)
    with get_tmpdir() as tmpdirname:
        # 2. convert the score to MEI usiing music21
        # TODO: if music21 implements export to MEI, use it
        midi_path = os.path.abspath(os.path.join(tmpdirname, "score.midi"))
        try:
            with open(os.devnull, 'w') as devnull:
                with contextlib.redirect_stdout(devnull):
                    score_data['score'].write("MIDI", midi_path)
        except Exception as e:
            filename = score_data[C.DATA_FILE]
            pwarn(f"jSymbolic: could not convert {filename} to MIDI: {e}")
            return

        # 3. run the MEI file through the jSymbolic jar savign csv into the temporary
        # directory in RAM
        out_path = os.path.abspath(os.path.join(tmpdirname, "features"))
        cmd = [JAVA_PATH,
               f"-Xmx{cfg.jsymbolic_max_ram}",
               "-jar",
               JSYMBOLIC_JAR,
               "-csv",
               ]
        if cfg.jsymbolic_config_file is not None:
            cmd += ["-configrun", cfg.jsymbolic_config_file]
        try:
            subprocess.run(
                cmd + [
                    midi_path,
                    out_path + ".xml",
                    out_path + "_def.xml",
                ],
                check=True,
                stdout=subprocess.DEVNULL,
            )
        except Exception as e:
            filename = score_data[C.DATA_FILE]
            pwarn(f"jSymbolic: cannot run jSymbolic on {filename}: {e}")
            return
        # 4. read the csv file into a pandas dataframe
        df = pd.read_csv(out_path + ".csv", na_values=["NaN", " NaN", "NaN ", " NaN "])
        df = df.drop(columns=df.columns[0])
        # 5. add `js_` prefix to the column names
        df.columns = ["js_" + c for c in df.columns]
        # 6. load the features into the score_features dictionary
        score_features.update(df.to_dict(orient='records')[0])


def update_part_objects(
    score_data: dict, part_data: dict, cfg: ExtractConfiguration, part_features: dict
):
    pass
